#!/usr/bin/env python3

##import dependency libraries
import sys,time,glob,os,pickle,fileinput
from subprocess import Popen
from operator import itemgetter
import gc, fileinput
import numpy as np
import argparse
from multiprocessing.dummy import Pool as ThreadPool
import multiprocessing
from copy import deepcopy
import ray
try:
    from pyminer.common_functions import *
except:
    from common_functions import *
#import pandas as pd


##############################################################

##########################################################################
parser = argparse.ArgumentParser()

## global arguments
parser.add_argument(
    '-infile','-in','-i','-input',
    dest='infile',
    type=str)

parser.add_argument(
    "-sample_groups",
    help='if you know how which samples belong to which groups, feed in a file that has the samples in the first column, and their group number (index starting at 0), in the second column. The IDs must be in the same order as in the infile too.',
    dest = 'sample_groups',
    type = str)

parser.add_argument(
    "-classes",
    help='if there are classes to compare, put the annotation table in this argument',
    dest = 'class_file',
    type = str)

parser.add_argument(
    "-within_group",
    help='if you know how which samples belong to which groups, feed in a file that has the samples in the first column, and their group number (index starting at 0), in the second column. The IDs must be in the same order as in the infile too.',
    dest = 'within_group',
    type = int)

parser.add_argument(
    "-out_dir","-o","-out",
    help='if you know how which samples belong to which groups, feed in a file that has the samples in the first column, and their group number (index starting at 0), in the second column. The IDs must be in the same order as in the infile too.',
    dest = 'out_dir',
    type = str)

parser.add_argument(
    "-species","-s",
    help = 'what species is this? Must be gProfiler compatible.',
    dest = 'species',
    type = str,
    default = 'hsapiens')

parser.add_argument(
    '-no_gProfile',
    help = 'should we do the automated gprofiler results?',
    default = False,
    action = 'store_true')

parser.add_argument(
    "-FDR","-fdr","-fdr_cutoff",
    help='The desired Benjamini-Hochberg False Discovery Rate (FDR) for multiple comparisons correction (default = 0.05)',
    dest = 'FDR_cutoff',
    type = float,
    default = 0.05)

parser.add_argument(
    "-Zscore","-Z_score_cutoff","-Z","-zscore","-z",
    help='The desired False Discovery Rate (FDR) for multiple comparisons correction (default = 0.05)',
    dest = 'Zscore',
    type = float,
    default = 2.0)

parser.add_argument(
    '-hdf5',
    help = 'The input file is an HDF5 file',
    default = False,
    action = 'store_true')

parser.add_argument(
    "-ID_list","-ids",
    help = 'If we are using an hdf5 file, give the row-wise IDs in this new line delimeted file',
    type = str)

parser.add_argument(
    "-columns","-cols",
    help = 'If we are using an hdf5 file, give the column-wise IDs in this new line delimeted file',
    type = str)

parser.add_argument(
    '-rows',
    help = 'if the samples are in rows, and variables are in columns',
    default = False,
    action = 'store_true')

parser.add_argument(
    "-log",'-log2','-log_transform',
    help='do a log transformation prior to clustering',
    action = 'store_true',
    default = False)

parser.add_argument(
    '-lin_norm',
    help = 'should we normalize the rows before doing the stats?',
    default = False,
    action = 'store_true')

parser.add_argument(
    '-processes', '-p',
    help = 'The number of processes to use. Default will be the number of available threads.',
    default = None,
    type = int)

args = parser.parse_args()
##########################################################################
args.infile = os.path.realpath(args.infile)
if args.out_dir == None:
    args.out_dir = get_file_path(args.infile)+'sample_clustering_and_summary/'
if args.out_dir[-1]!='/':
    args.out_dir+='/'

sample_dir = args.out_dir
temp = args.out_dir
cmd('mkdir '+args.out_dir)

##############################################################################
## read in the dataset and sample groups

if args.within_group==None:
    sample_group_table = read_table(args.sample_groups)

    sample_group_table_np = np.array(sample_group_table)
    sample_group_order = np.transpose(sample_group_table_np[:,0])
    sorted_list_of_ids = list(sample_group_order)

    grouping_vector = list(np.transpose(sample_group_table_np[:,1]))
    #print(grouping_vector)
    #sys.exit()
else:
    if args.class_file==None:
        sys.exit("to run within class analysis, we need the classes annotation file: -classes")
    orig_sample_group_table = read_table(args.sample_groups)
    orig_sample_group_table_np = np.array(orig_sample_group_table)
    orig_sample_group_order = np.transpose(orig_sample_group_table_np[:,0])
    orig_sorted_list_of_ids = list(orig_sample_group_order)
    orig_grouping_vector = list(np.transpose(orig_sample_group_table_np[:,1]))
    ## we're going to switch around the normal sample groups with the 
    ## classes that are read in 
    class_table_np = np.array(read_table(args.class_file))
    class_labels = list(set(class_table_np[:,1].tolist()))
    class_labels = sorted(class_labels)
    class_hash = {key:value for value, key in enumerate(class_labels)}
    alias_dict = {key:value for key, value in enumerate(class_labels)}
    ## make a table to define which class is which group
    class_index_table = []
    for c in class_labels:
        class_index_table.append([c,class_hash[c]])
        #print(class_index_table[-1])
    ## figure out which indices to keep
    subset_col_idxs=[]
    for i in range(0,len(orig_grouping_vector)):
        if str(orig_grouping_vector[i])==str(float(args.within_group)):
            subset_col_idxs.append(i)
        #print(orig_grouping_vector[i])
    #print(subset_col_idxs,float(args.within_group))
    sample_group_table_np = class_table_np[subset_col_idxs]
    ## go through the annotation table and make
    for i in range(len(subset_col_idxs)):
        sample_group_table_np[i,1]=class_hash[sample_group_table_np[i,1]]
    #print(sample_group_table_np)

    sample_group_table = sample_group_table_np.tolist()
    for i in range(len(sample_group_table)):
        sample_group_table[i][1]=float(sample_group_table[i][1])
    sample_group_table_np = np.array(sample_group_table)
    sample_group_order = np.transpose(sample_group_table_np[:,0])
    sorted_list_of_ids = list(sample_group_order)
    grouping_vector = list(np.transpose(sample_group_table_np[:,1]))
    subset_col_idxs=np.array(subset_col_idxs,dtype=int)
    #print(grouping_vector)
    


if len(list(set(grouping_vector))) == 1:
    one_group = True
    sys.exit('only one sample group, nothing to calculate')
else:
    one_group = False




sample_cluster_ids = []
for i in range(0,len(sample_group_table)):
    
    ## THIS IS IMPORTANT
    ## here we assume that the samples are all listed in the same order as in '-infile'
    ## we also assume that the group indexing starts at 0
    sample_cluster_ids.append(sample_group_table[i][1])
sample_cluster_ids = list(map(int,sample_cluster_ids))
sample_k_lists = []
for i in range(0,max(sample_cluster_ids)+1):
    sample_k_lists.append([])
#print(len(sample_k_lists))
## now populate the list of lists
for i in range(0,len(sample_cluster_ids)):
    ## this appends the sample index to 
    #print(sample_cluster_ids[i])
    sample_k_lists[sample_cluster_ids[i]].append(i)

#print(sample_k_lists)




## check for formatting
if not args.hdf5:
    full_expression_str = read_table(args.infile)
    title = full_expression_str[0]
    full_expression_np = np.array(full_expression_str)
    row_names = full_expression_np[1:,0]
    full_expression = np.array(full_expression_np[1:,1:],dtype = float)
else:
    row_names = read_file(args.ID_list,'lines')
    title = read_file(args.columns,'lines')
    print('making a maliable hdf5 file to preserve the original data')
    cmd('cp '+args.infile+' '+args.infile+'_copy')
    import h5py
    print('reading in hdf5 file')
    infile_path = args.infile+'_copy'
    h5f = h5py.File(infile_path, 'r+')
    full_expression=h5f["infile"]

##########################################################
## if we're doing the subset analysis, subset which columns to include
if args.within_group!=None:
    full_expression = full_expression[:,subset_col_idxs]
    temp_title = np.array(title[1:])
    title = [title[0]] + temp_title[subset_col_idxs].tolist()
    print(row_names)

##########################################################
if args.rows:
    full_expression = np.transpose(full_expression)
    temp_col = ["variables"] + row_names.tolist()
    title = temp_col[:]


if title[1:] != sorted_list_of_ids:
    print('length of columns',len(title[1:]))
    print('length of sample_groups',sorted_list_of_ids)
    for i in range(1,len(title)):
        if title[i] != sorted_list_of_ids[i-1]:
            print('on line',i,'ID_list',title[i],'',sorted_list_of_ids[i-1])
    sys.exit("the sample group ids have to be in the same order as in the input matrix")



def lin_norm_rows(in_mat,min_range=0,max_range=1):
    in_mat = np.transpose(np.array(in_mat))
    in_mat = in_mat - np.min(in_mat, axis = 0)
    in_mat = in_mat / np.max(in_mat, axis = 0)
    in_mat[np.isnan(in_mat)]=0
    return(np.transpose(in_mat))


## make all of the nans equal to zero
full_expression[np.isnan(full_expression)] = 0.0
if args.log:
    full_expression = np.log2(full_expression-np.min(full_expression)+1)
if args.lin_norm:
    full_expression = lin_norm_rows(full_expression)

IDlist = list(row_names)

##############################################################################
## calculate the sample level Z-scores


## first normalize the expression matrix linearly between 1 and 2
#sample_var_enrichment = lin_norm_rows(full_expression) + 1
if not args.hdf5:
    sample_var_enrichment = full_expression

    ## then calculate the mean for each variable
    norm_row_means = np.transpose(np.array([np.mean(sample_var_enrichment, axis = 1)]))
    #print(norm_row_means)
    ## then calculate the sd for each variable
    norm_row_sd = np.transpose(np.array([np.std(sample_var_enrichment, axis = 1)]))
    #print(norm_row_sd)
    ## then calculate the delta for each variable
    norm_row_delta = sample_var_enrichment - norm_row_means

    ## calculate the z-score (ie: how many standard deviations away from the mean is each sample)
    sample_var_enrichment_numeric = norm_row_delta/norm_row_sd
    #print(sample_var_enrichment_numeric)
    #sys.exit()



    ## add the titles

    sample_var_enrichment = np.array([title[1:]] + sample_var_enrichment_numeric.tolist()).tolist()

    temp_ids = ['variables']+IDlist
    for i in range(0,len(sample_var_enrichment)):
        sample_var_enrichment[i]=[temp_ids[i]]+sample_var_enrichment[i]

    #sample_var_enrichment = np.hstack((row_titles,sample_var_enrichment))

    write_table(sample_var_enrichment, sample_dir+'sample_var_enrichment_Zscores.txt')




##############################################################################
## go through each group and generate mean and sd summaries, then write to file

################################
## ray functions
def get_indices(threads, num_genes):
    indices_list = []
    for t in range(threads):
        indices_list.append([])
    temp_idx = 0
    while temp_idx < num_genes:
        for t in range(threads):
            if temp_idx < num_genes:
                indices_list[t].append(temp_idx)
                temp_idx += 1
    return(indices_list)


def get_num_rows_from_dict_lists(dict_list):
    ## goes through all of the indices & returns the number of dims
    row_dims = 0
    for temp_dict in dict_list:
        for temp_key in list(temp_dict.keys()):
            #print(temp_key)
            if temp_key > row_dims:
                row_dims = temp_key
    return(row_dims+1)


def get_num_cols_from_dict_lists(dict_list):
    ## first get the dimentions
    first_key = list(dict_list[0].keys())[0]
    first = dict_list[0][first_key]
    if type(first) == list:
        col_dims = len(first)
    else:
        col_dims = 1
    return(col_dims)


def ray_dicts_to_array(dict_list):
    row_dims = get_num_rows_from_dict_lists(dict_list)
    col_dims = get_num_cols_from_dict_lists(dict_list)
    out_array = np.zeros((row_dims, col_dims))
    print(out_array)
    for temp_dict in dict_list:
        for idx, value in temp_dict.items():
            #print(idx)
            #print(value)
            out_array[idx] = value
    return(out_array)


@ray.remote
def ray_get_global_express_stats(indices, 
                                 ray_full_expression,
                                 hdf5_file = None):
    ## load the ray_full_express if it's hdf5
    if hdf5_file is not None:
        h5f = h5py.File(hdf5_file, 'r')
        ray_full_expression=h5f["infile"]

    global_percent_express_dict = {index:None for index in indices}
    global_non_zero_mean_dict = {index:None for index in indices}

    ray_count = 0

    for index in indices:
        if ray_count % 250 == 0:
            print("worker is:",ray_count/len(indices)*100," percent done")
        temp_binary = np.array(ray_full_expression[index,:] > 0,dtype=bool)
        temp_binary_int = np.array(deepcopy(temp_binary),dtype=int)
        temp_num_express = np.sum(temp_binary_int)
        global_percent_express = temp_num_express/temp_binary_int.shape[0]
        global_non_zero_mean = np.sum(ray_full_expression[index,:])/temp_num_express
        global_percent_express_dict[index] = global_percent_express
        global_non_zero_mean_dict[index] = global_non_zero_mean

        ray_count+=1

    if hdf5_file is not None:
        h5f.close()

    return(global_percent_express_dict, global_non_zero_mean_dict)


@ray.remote
def ray_get_cell_type_express_stats(indices,
                                    ray_full_expression,
                                    ray_list_of_k_sample_indices,
                                    hdf5_file = None):
    ## load the ray_full_express if it's hdf5
    if hdf5_file is not None:
        h5f = h5py.File(hdf5_file, 'r')
        ray_full_expression=h5f["infile"]

    group_percent_express_dict = {index:None for index in indices}
    group_non_zero_mean_dict = {index:None for index in indices}
    
    ray_count = 0
    
    for index in indices:
        if ray_count % 250 == 0:
            print("worker is:",ray_count/len(indices)*100," percent done")

        group_percent_express = []
        group_non_zero_mean = []
        for i in range(len(list_of_k_sample_indices)):
            temp_binary = np.array(ray_full_expression[index,list_of_k_sample_indices[i]] > 0,dtype=bool)
            temp_binary_int = np.array(deepcopy(temp_binary),dtype=int)
            temp_num_express = np.sum(temp_binary_int)
            temp_percent_express = temp_num_express/len(list_of_k_sample_indices[i])
            temp_non_zero_mean = np.sum(ray_full_expression[index,list_of_k_sample_indices[i]])/temp_num_express
            group_percent_express.append(temp_percent_express)
            group_non_zero_mean.append(temp_non_zero_mean)

        group_percent_express_dict[index] = group_percent_express
        group_non_zero_mean_dict[index] = group_non_zero_mean
        
        ray_count += 1
    
    if hdf5_file is not None:
        h5f.close()
    return(group_percent_express_dict, group_non_zero_mean_dict)

#####
## anova functions

@ray.remote
def ray_get_anova(indices,
                  ray_full_expression,
                  ray_list_of_k_sample_indices,
                  hdf5_file = None):
    ## load the ray_full_express if it's hdf5
    if hdf5_file is not None:
        h5f = h5py.File(hdf5_file, 'r')
        ray_full_expression=h5f["infile"]

    aov_dict = {index:None for index in indices}

    ray_count = 0
    for index in indices:
        if ray_count % 250 == 0:
            print("worker is:",ray_count/len(indices)*100," percent done")
        if np.sum(ray_full_expression[index,:]) == 0:
            aov_dict[index] = [np.nan,np.nan]
        list_of_group_values=[]
        for group in range(0,len(ray_list_of_k_sample_indices)):
            #print(len(list_of_k_sample_indices[group]))
            list_of_group_values.append(ray_full_expression[index,ray_list_of_k_sample_indices[group]])
            #print(np.shape(list_of_group_values[-1]))
            #print(list_of_group_values[-1])
            
        return_val = list(aov(*list_of_group_values))
        aov_dict[index] = return_val

        ray_count += 1

    if hdf5_file is not None:
        h5f.close()

    return(aov_dict)











## /ray functions
################################


list_of_k_sample_indices = sample_k_lists
#print(list_of_k_sample_indices)

if not one_group:
    
    sample_names = list(title)[1:]
    #print(sample_names)
    # list_of_k_sample_indices=[]
    
    # for k in sample_k_lists:
    #     print(k)
    #     temp_indices_list=[]
    #     for i in range(0,len(k)):
    #         temp_indices_list.append(sample_names.index(k[i]))
    #     list_of_k_sample_indices.append(temp_indices_list)


    #print(list_of_k_sample_indices)
    list_of_k_sample_indices = np.array(list_of_k_sample_indices)

    ## initialize mean and sd output tables
    #print(list_of_k_sample_indices[0])
    #test=full_expression[:,list_of_k_sample_indices[0]]
    #print(np.mean(test.astype('float32'), axis = 1))
    #print(full_expression[:,list_of_k_sample_indices[0]].shape)



    ## calculate the means accross rows
    all_sample_mean = np.transpose(np.array([np.mean(full_expression, axis=1)]))
    all_sample_sd = np.transpose(np.array([np.std(full_expression, axis=1)]))
    k_group_means = np.transpose(np.array([np.mean(full_expression[:,list_of_k_sample_indices[0]], axis=1)]))
    k_group_sd = np.transpose(np.array([np.std(full_expression[:,list_of_k_sample_indices[0]], axis=1)]))

    for k in range(1,len(list_of_k_sample_indices)):## start from 1 because we already initiated it on the first col
        ## calc row means
        new_mean_col = np.transpose(np.array([np.mean(full_expression[:,list_of_k_sample_indices[k]], axis=1)]))
        k_group_means = np.hstack((k_group_means, new_mean_col))
    
        ## calc sd
        new_sd_col = np.transpose(np.array([np.std(full_expression[:,list_of_k_sample_indices[k]], axis=1)]))
        k_group_sd = np.hstack((k_group_sd, new_sd_col))

    ###########################################################
    ## TODO
    ## calculate the percent non-zero for each cell type
    print("calculating global percent expressed and mean non-zero expression")
    k_group_percent_express = np.zeros((full_expression.shape[0], len(list_of_k_sample_indices))) ## the percentage 
    ratio_percent_express = np.zeros((full_expression.shape[0], len(list_of_k_sample_indices))) ## ratio of within cell type percent expression to global percent expression
    non_zero_mean_expression = np.zeros((full_expression.shape[0], len(list_of_k_sample_indices))) ## ratio of within cell type percent expression to global percent expression
    ratio_non_zero_mean_expression = np.zeros((full_expression.shape[0], len(list_of_k_sample_indices))) ## ratio of within cell type percent expression to global percent expression
    
    # #######################################################################################
    # ################## OLD ##############################
    # #####################################################
    # ## first do the global expression calculations
    if args.processes == 1:
        def get_global_express_stats(index):
            global full_expression
            temp_binary = np.array(full_expression[index,:] > 0,dtype=bool)
            temp_binary_int = np.array(deepcopy(temp_binary),dtype=int)
            temp_num_express = np.sum(temp_binary_int)
            global_percent_express = temp_num_express/temp_binary_int.shape[0]
            global_non_zero_mean = np.sum(full_expression[index,:])/temp_num_express
            return(global_percent_express, global_non_zero_mean)
        
        ## set up multi-threaded global stats
        ## global_percent_express, global_non_zero_mean
        threads = multiprocessing.cpu_count()
        pool = ThreadPool(threads)
        global_results = pool.map(get_global_express_stats,range(full_expression.shape[0]))
        pool.close()
        pool.join()

        ## separate results
        global_percent_express = []
        global_non_zero_mean = []
        for i in range(len(global_results)):
            global_percent_express.append(global_results[i][0])
            global_non_zero_mean.append(global_results[i][1])

        del global_results
        global_percent_express = np.array(global_percent_express)
        global_non_zero_mean = np.array(global_non_zero_mean)

        # ######################## OLD #####################################
        ## get cell type level results
        def get_cell_type_express_stats(index):
            global full_expression, list_of_k_sample_indices
            group_percent_express = []
            group_non_zero_mean = []
            for i in range(len(list_of_k_sample_indices)):
                temp_binary = np.array(full_expression[index,list_of_k_sample_indices[i]] > 0,dtype=bool)
                temp_binary_int = np.array(deepcopy(temp_binary),dtype=int)
                temp_num_express = np.sum(temp_binary_int)
                temp_percent_express = temp_num_express/len(list_of_k_sample_indices[i])
                temp_non_zero_mean = np.sum(full_expression[index,list_of_k_sample_indices[i]])/temp_num_express
                group_percent_express.append(temp_percent_express)
                group_non_zero_mean.append(temp_non_zero_mean)
            return(group_percent_express, group_non_zero_mean)


        ###########################################################
        ## set up multi-threaded global stats
        ## global_percent_express, global_non_zero_mean
        threads = multiprocessing.cpu_count()
        pool = ThreadPool(threads)
        group_results = pool.map(get_cell_type_express_stats,range(full_expression.shape[0]))
        pool.close()
        pool.join()
        
        for i in range(len(group_results)):
            k_group_percent_express[i,:] = group_results[i][0]
            non_zero_mean_expression[i,:] = group_results[i][1]
        del group_results

        print(k_group_percent_express)
        print(non_zero_mean_expression)
        

        ######################## /OLD #####################################

    #################### /OLD ###########################
    #################### NEW ############################
    else:
        if args.processes == None:
            args.processes = multiprocessing.cpu_count()
        threads = args.processes
        indices_list = get_indices(threads, full_expression.shape[0])
        
        ray.init()

        ###########
        if not args.hdf5:
            ray_full_expression = ray.put(full_expression)
        else:
            ## make copies of the input hdf5 file
            hdf5_file_list = []
            for t in range(threads):
                hdf5_file_list.append(args.infile+"_"+str(t))
                cmd('cp '+args.infile+" "+hdf5_file_list[-1])
        ###########

        r_jobs = []
        for t in range(threads):
            if not args.hdf5:
                r_jobs.append(ray_get_global_express_stats.remote(indices_list[t], 
                                                                  ray_full_expression))
            else:
                r_jobs.append(ray_get_global_express_stats.remote(indices_list[t], 
                                                                  None,
                                                                  hdf5_file = hdf5_file_list[t]))
        temp_r_results = ray.get(r_jobs)
        ray.shutdown()
    
        global_percent_express_dict_list = []
        global_non_zero_mean_dict_list = []

        for t in range(len(temp_r_results)):
            global_percent_express_dict_list.append(temp_r_results[t][0])
            global_non_zero_mean_dict_list.append(temp_r_results[t][1])

        global_percent_express = ray_dicts_to_array(global_percent_express_dict_list)
        global_non_zero_mean = ray_dicts_to_array(global_non_zero_mean_dict_list)

        global_percent_express = global_percent_express[:,0]
        global_non_zero_mean = global_non_zero_mean[:,0]

        print(global_percent_express)
        print(global_non_zero_mean)
        #################### /NEW ############################    
        #######################################################################################

        print("calculating percent expressed in cell types and mean non-zero expression")

        
        ######################## NEW #####################################
        ray.init()
        if not args.hdf5:
            ray_full_expression = ray.put(full_expression)

        ###########
        # indices,
        # ray_full_expression,
        # ray_list_of_k_sample_indices,
        # hdf5_file = None)

        r_jobs = []

        if args.processes == None:
            args.processes = multiprocessing.cpu_count()
        threads = args.processes
        for t in range(threads):
            if not args.hdf5:
                r_jobs.append(ray_get_cell_type_express_stats.remote(indices_list[t], 
                                                                     ray_full_expression,
                                                                     list_of_k_sample_indices))
            else:
                r_jobs.append(ray_get_cell_type_express_stats.remote(indices_list[t], 
                                                                     None,
                                                                     list_of_k_sample_indices,
                                                                     hdf5_file = hdf5_file_list[t]))
        temp_r_results = ray.get(r_jobs)
        ray.shutdown()
        
        k_group_percent_express_dict_list = []
        non_zero_mean_expression_dict_list = []

        for t in range(len(temp_r_results)):
            k_group_percent_express_dict_list.append(temp_r_results[t][0])
            non_zero_mean_expression_dict_list.append(temp_r_results[t][1])

        k_group_percent_express = ray_dicts_to_array(k_group_percent_express_dict_list)
        non_zero_mean_expression = ray_dicts_to_array(non_zero_mean_expression_dict_list)

        print(k_group_percent_express)
        print(non_zero_mean_expression)

        ######################## /NEW #####################################

    ## calculate ratios
    print("global_percent_express",global_percent_express.shape)
    print("global_non_zero_mean",global_non_zero_mean.shape)
    print("k_group_percent_express", k_group_percent_express.shape)
    print("non_zero_mean_expression", non_zero_mean_expression.shape)
    print("ratio_percent_express",ratio_percent_express.shape)
    print("ratio_non_zero_mean_expression",ratio_non_zero_mean_expression.shape)
    for i in range(k_group_percent_express.shape[1]):
        ratio_percent_express[:,i] = k_group_percent_express[:,i]/global_percent_express
        ratio_non_zero_mean_expression[:,i] = non_zero_mean_expression[:,i]/global_non_zero_mean

    
##############################################################################
########### use the means of the k-means groups to calculate the enrichment



    ## calculate the sample group z-scores ((Xbar-Mu)/sigma) * sqrt(n), or (Xbar-Mu)/(sigma/sqrt(n))
    print("calculating the sample group vs all sample means")
    row_delta = np.array(k_group_means.transpose() - all_sample_mean.transpose()).transpose()
    print(np.shape(row_delta))
    for k in range(0,len(list_of_k_sample_indices)):
        #print(k)
        row_delta[:,k] = np.array(row_delta[:,k].transpose() / all_sample_sd[:,0].transpose()).transpose()
        row_delta[:,k] = row_delta[:,k] * np.sqrt(len(list_of_k_sample_indices[k]))

    sample_k_group_enrichment = row_delta
    sample_k_group_enrichment_numeric = sample_k_group_enrichment

    #######################################################

    sample_groups = []
    for k in range(0,len(list_of_k_sample_indices)):
        #print(k)
        if 'alias_dict' in globals():
            #print(list(alias_dict.keys()))
            temp_sample_group_name = alias_dict[k]
        else:
            temp_sample_group_name = 'sample_group_'+str(k)
        sample_groups.append(temp_sample_group_name)
        #print(sample_groups)
    sample_groups = np.array(sample_groups)


    ## get ready to write it to file
    k_group_means = np.vstack((sample_groups,k_group_means))
    k_group_sd = np.vstack((sample_groups,k_group_sd))
    sample_k_group_enrichment = np.vstack((sample_groups, sample_k_group_enrichment))

    global_percent_express = np.array(["global_percent_express"]+global_percent_express.tolist())
    global_non_zero_mean = np.array(["global_non_zero_mean"]+global_non_zero_mean.tolist())

    k_group_percent_express = np.vstack((sample_groups, k_group_percent_express))
    non_zero_mean_expression = np.vstack((sample_groups, non_zero_mean_expression))
    ratio_percent_express = np.vstack((sample_groups, ratio_percent_express))

    ratio_non_zero_mean_expression = np.vstack((sample_groups, ratio_non_zero_mean_expression))

    ##############################################################################
    ##### prepare the k-means enrichment for writing to file, then write it
    
    ## these are the names of the variables, making the row labels for the output table
    row_names = np.transpose(np.array([['var_names']+IDlist]))
    
    
    k_group_means = np.hstack((row_names, k_group_means))
    k_group_sd = np.hstack((row_names, k_group_sd))
    sample_k_group_enrichment = np.hstack((row_names, sample_k_group_enrichment))

    k_group_percent_express = np.hstack((row_names, k_group_percent_express))
    non_zero_mean_expression = np.hstack((row_names, non_zero_mean_expression))
    ratio_percent_express = np.hstack((row_names, ratio_percent_express))
    ratio_non_zero_mean_expression = np.hstack((row_names, ratio_non_zero_mean_expression))

    global_percent_express = list(map(list,zip(['var_names']+IDlist, global_percent_express.tolist())))
    global_non_zero_mean = list(map(list,zip(['var_names']+IDlist, global_non_zero_mean.tolist())))


    write_table(k_group_means,sample_dir+'k_group_means.txt')
    write_table(k_group_sd,sample_dir+'k_group_sd.txt')
    write_table(sample_k_group_enrichment, sample_dir+'k_group_enrichment.txt')

    express_metrics_dir = process_dir(os.path.join(sample_dir,'expression_metrics'))
    write_table(global_percent_express,os.path.join(express_metrics_dir,"global_percent_express.tsv"))
    write_table(global_non_zero_mean,os.path.join(express_metrics_dir,"global_non_zero_mean.tsv"))
    write_table(k_group_percent_express,os.path.join(express_metrics_dir,"k_group_percent_express.tsv"))
    write_table(non_zero_mean_expression,os.path.join(express_metrics_dir,"non_zero_mean_expression.tsv"))
    write_table(ratio_percent_express,os.path.join(express_metrics_dir,"ratio_percent_express.tsv"))
    write_table(ratio_non_zero_mean_expression,os.path.join(express_metrics_dir,"ratio_non_zero_mean_expression.tsv"))


##################################################################################################################
##### do a one way anova between groups to test for variables that are significantly different between them ######

from scipy.stats.mstats import f_oneway as aov
from scipy.stats import ttest_ind as ttest

def get_ttests(index, aov_p):
    ## for use in the context of a Friedman's Protected LSD post-hoc to significant anovas
    global args, full_expression, list_of_k_sample_indices
    p_vals = np.ones((len(list_of_k_sample_indices),len(list_of_k_sample_indices)))
    if aov_p > args.FDR_cutoff:
        return(p_vals)
    list_of_group_values=[]
    for group in range(0,len(list_of_k_sample_indices)):
        list_of_group_values.append(full_expression[index,list_of_k_sample_indices[group]])
    for i in range(0,len(list_of_group_values)-1):
        for j in range(i+1, len(list_of_group_values)):
            stat, p = ttest(list_of_group_values[i],list_of_group_values[j])
            p_vals[i,j]=p
            p_vals[j,i]=p
    return(p_vals)

def get_anova(index):
    #print('\n'*10)
    global full_expression, list_of_k_sample_indices
    if np.sum(full_expression[index,:]) == 0:
        return([np.nan,np.nan])
    list_of_group_values=[]
    for group in range(0,len(list_of_k_sample_indices)):
        #print(len(list_of_k_sample_indices[group]))
        list_of_group_values.append(full_expression[index,list_of_k_sample_indices[group]])
        #print(np.shape(list_of_group_values[-1]))
        #print(list_of_group_values[-1])
        
    return_val = list(aov(*list_of_group_values))
    return(return_val)

# if not very_big_file:
#     if len(list_of_k_sample_indices) > 1:
#         enough_k_samples = True
#     else:
#         enough_k_samples = False
# else:
#     enough_k_samples = False

all_sig_enriched_files=[]
all_sig_enriched_gprof_files = []
FDR_cutoff = args.FDR_cutoff
zscore_cutoff = args.Zscore
if not one_group:
    anova_output = [['Variable', 'F', 'uncorrected_p-val', 'BH_corrected_p-val','-log10(BH_cor_p-val)']]
    aov_uncorrected_p_val_list=[]

    if args.processes==1:
        ######### OLD ######################
        # go through each variable and get the uncorrected anova stats
        threads = multiprocessing.cpu_count()
        print('doing the 1-way ANOVAs:',threads,"threads")
        pool = ThreadPool(threads)
        all_aov_results = pool.map(get_anova,range(full_expression.shape[0]))
        pool.close()
        pool.join()
        print('done')

        print(all_aov_results[:5])
        ########## /OLD ######################
    else:
        ######### NEW ##################
        ray.init()
        if not args.hdf5:
            ray_full_expression = ray.put(full_expression)
        ###########
        # indices,
        # ray_full_expression,
        # ray_list_of_k_sample_indices,
        # hdf5_file = None)

        r_jobs = []

        for t in range(threads):
            if not args.hdf5:
                r_jobs.append(ray_get_anova.remote(indices_list[t],
                                                   ray_full_expression,
                                                   list_of_k_sample_indices))
            else:
                r_jobs.append(ray_get_anova.remote(indices_list[t], 
                                                   None,
                                                   list_of_k_sample_indices,
                                                   hdf5_file = hdf5_file_list[t]))


        temp_aov_results = ray.get(r_jobs)
        ray.shutdown()
            
        all_aov_results = ray_dicts_to_array(temp_aov_results).tolist()
        
        ## clean up the copied files
        if args.hdf5:
            for file in hdf5_file_list:
                cmd('rm '+file)

        ######### /NEW ##################
    #################################


    ## now do do the BH correction
    for i in range(0,len(IDlist)):
        anova_output.append([IDlist[i]]+all_aov_results[i])
        aov_uncorrected_p_val_list.append(anova_output[-1][-1])

    ## replace nans with 1s
    aov_uncorrected_p_val_list=np.array(aov_uncorrected_p_val_list)
    aov_uncorrected_p_val_list[np.isnan(aov_uncorrected_p_val_list)] = 1

    #print(aov_uncorrected_p_val_list[:10])
    #print(correct_pvalues_for_multiple_testing(aov_uncorrected_p_val_list[:10]))
    #for line in anova_output:
    #    print(line)
    ## correct the p-values with Benjamini-Hochberg 
    BH_corrected_aov = correct_pvalues_for_multiple_testing(aov_uncorrected_p_val_list)
    #print(BH_corrected_aov[:10])

    indices_less_than_FDR_cutoff=[]
    for i in range(0,len(IDlist)):
        anova_output[i+1]+=[BH_corrected_aov[i], -1*np.log10(BH_corrected_aov[i])]
        if BH_corrected_aov[i] <= FDR_cutoff:
            indices_less_than_FDR_cutoff.append(i)


    #cmd('mkdir "'+args.out_dir+'significance"')
    process_dir(os.path.join(args.out_dir,'significance'))
    write_table(anova_output , args.out_dir+'significance/groups_1way_anova_results.txt')

    ################################################################################################
    ############# Find the significantly different & enriched variables for each group #############
    ################################################################################################

    ## we start off with the variables that are significantly different between groups
    
    ## this is a list of lists for each 
    group_sig_enriched_bool=[]


    for i in range(0,len(IDlist)):
        ## check to see if this variable was significant in the BH corrected 1-way anova
        if i in indices_less_than_FDR_cutoff:
            var_significant = True
        else:
            var_significant = False
        
        ## if it's significant, we want to go in and check which, if any sample groups 
        ## show an elevated z-score enrichment (greater than the zscore_cutoff variable)
        if var_significant:
            #print(sample_k_group_enrichment_numeric[i,])
            temp_sig_enriched_bool_list = list(sample_k_group_enrichment_numeric[i,] >= zscore_cutoff)
        else: 
            temp_sig_enriched_bool_list = [False]*len(list_of_k_sample_indices)
        
        group_sig_enriched_bool.append(temp_sig_enriched_bool_list)
    
    bool_sample_groups = []
    for s in sample_groups:
        bool_sample_groups.append('bool_'+s)
    bool_sample_groups = np.array(bool_sample_groups)
    
    group_sig_enriched_bool = np.array(group_sig_enriched_bool)
    group_sig_enriched = np.vstack((bool_sample_groups,group_sig_enriched_bool))
    group_sig_enriched = np.hstack((row_names,group_sig_enriched))
    
    write_table(group_sig_enriched, args.out_dir+'significance/significant_and_enriched_boolean_table.txt')
    
    ## go through each group, and make a list of the variables that are significant and enriched
    num_sig_enriched = []
    for group in range(0,len(list_of_k_sample_indices)):
        sig_and_enriched_for_this_group = []
        for var in range(0,len(IDlist)):
            if group_sig_enriched_bool[var,group]:
                sig_and_enriched_for_this_group.append(IDlist[var])
        num_sig_enriched.append(len(sig_and_enriched_for_this_group))
        temp_file = args.out_dir+'significance/'+sample_groups[group]+'_significant_enriched.txt'
        all_sig_enriched_files.append(temp_file)
        all_sig_enriched_gprof_files.append(args.out_dir+'significance/gprofiler/'+sample_groups[group]+'_significant_enriched_gprofiler.txt')
        make_file('\n'.join(sig_and_enriched_for_this_group), temp_file)
    
#    print(np.shape(sample_groups))
#    print(np.shape(np.array(num_sig_enriched)))
#    print(num_sig_enriched)
    num_sig_enriched = np.transpose(np.vstack((np.array([sample_groups]),np.array([num_sig_enriched]))))
    
    num_sig_enriched = np.vstack((np.array([['sample_group', 'number_of_significant_and_enriched_vars']]), num_sig_enriched))
    
    write_table(num_sig_enriched, args.out_dir+'significance/number_of_variables_significant_and_enriched_for_each_group.txt')


enriched_in_one_bool = np.sum(group_sig_enriched_bool, axis = 1) != 0
temp_IDs = np.array(IDlist)
IDs_enriched_in_one = temp_IDs[enriched_in_one_bool].tolist()
background = args.out_dir+'significance/background_ids.txt'
make_file('\n'.join(IDlist),background)

if not args.no_gProfile:
    cmd('mkdir '+args.out_dir+'significance/gprofiler/')

    for f in range(0,len(all_sig_enriched_files)):
        temp_gprofile = 'pyminer_gprofile.py -i '+all_sig_enriched_files[f]+' -s '+args.species+' -o '+all_sig_enriched_gprof_files[f]
        temp_gprofile += ' -b '+background
        cmd(temp_gprofile)















#!/usr/local/env python3

import argparse
#from requests_html import HTMLSession
#import urllib
import requests
import time
from bs4 import BeautifulSoup as bs
from copy import deepcopy
try:
    from pyminer.common_functions import *
except:
    from common_functions import *
##############################################################

def get_web_html(url, tries = 3, timeout = 0.5, use_requests = True):
    if requests:
        resp = requests.get(url)
        if str(resp) == "<Response [200]>":
            return(resp.text)
        else:
            print("couldn't connect!")
            time.sleep(.25)
            return(get_web_html(url,tries = tries -1, timeout = timeout, use_requests = use_requests))
    print("\t",url)
    if tries == 0:
        return("")
    #global session
    session = HTMLSession()
    html_text = ""
    # Use the object above to connect to needed webpage
    try:
        resp = session.get(url, timeout = timeout)
    except:
        print("\t\tcouldn't connect to the url")
        try:
            session.close()
        except:
            print("\t\tCouldn't close session!")
        return(get_web_html(url, tries = tries-1, timeout = timeout+0.25))
    else:
        try:
            resp.html.render()
        except:
            print("\t\tcouldn't render it!; we'll return what we got anyway")
            html_text = str(resp.html.html)
        else:
            html_text = str(resp.html.html)
    try:
        resp.close()
        session.close()
    except:
        print("\t\tCouldn't close session!")
    return(html_text)


def extract_counts(line):
    line = line.split(' ')
    line = line[-1]
    return(int(line))


def count_from_html(html):
    if html == "":
        return("")
    soup = bs(html,"lxml")
    p_list = soup.findAll('h3')
    counts = 1
    for p in p_list:
        #out_text += " . "+p.text
        #print(p.text)
        if "Items: " in p.text:
            counts = extract_counts(p.text)
            return(counts)
    return(counts)


def get_search_count(url):
    temp_html = get_web_html(url)
    count = count_from_html(temp_html)
    return(count)

def prep_url(term1,term2='', no_search_quote = False, base = 'https://www.ncbi.nlm.nih.gov/pubmed/?term='):
    term1 = term1.split(' ')
    term1 = '"'+"+".join(term1)+'"'
    search_term = term1
    if term2 != '':
        term2 = term2.split(' ')
        if not no_search_quote:
            term2 = '"'+"+".join(term2)+'"'
        else:
            term2 = "+".join(term2)
        search_term = term1+"+"+term2
    return(base+search_term+"+NOT+(GWAS)")


def count_from_html_new_pubmed(html):
    if html == "":
        return("")
    soup = bs(html,"lxml")
    p_list = soup.findAll('div', {'class':"results-amount"})
    counts = 1
    for p in p_list:
        values = p.findAll('span', {'class':"value"})
        if len(values) == 0:
            return(0)
        for value in values:
            value_text = value.get_text()
            value_text = value_text.replace(',','')
            #print(value_text)
            if "No results" in value_text:
                counts = 0
                return(counts)
            try:
                int(value_text)
            except:
                print("Something went wrong with:", value_text)
            else:
                counts = int(value_text)
                return(counts)
    return(counts)


def prep_new_pubmed_url(term1, term2='', no_search_quote = False):
    return(prep_url(term1, term2=term2, no_search_quote = no_search_quote, base = "https://pubmed.ncbi.nlm.nih.gov/?term="))


def get_new_pubmed_search_count(url):
    temp_html = get_web_html(url)
    count = count_from_html_new_pubmed(temp_html)
    return(count)


def get_single_search_count(term1, no_search_quote = False, new_pubmed = False):
    if term1 == "None":
        return(0)
    if not new_pubmed:
        term1_url = prep_url(term1, no_search_quote = no_search_quote)
        term1_count = get_search_count(term1_url)
    else:
        term1_url = prep_new_pubmed_url(term1, no_search_quote = no_search_quote)
        term1_count = get_new_pubmed_search_count(term1_url)
    print(term1,'\t',term1_count)
    return(term1_count)
    

def get_pubmed_comentions(term1, term2, no_search_quote = False, new_pubmed = False):
    if term1=="None" or term2 == "None":
        return(0)
    if not new_pubmed:
        combined_url = prep_url(term1, term2= term2, no_search_quote = no_search_quote)
        combined_count = get_search_count(combined_url)
    else:
        combined_url = prep_new_pubmed_url(term1, term2= term2, no_search_quote = no_search_quote)
        combined_count = get_new_pubmed_search_count(combined_url)
    print(term1, '\t', term2,'\t',combined_count)
    return(combined_count)


def get_all_result_comparisons(genes,search_term, no_search_quote = False, new_pubmed = False):
    gene_counts = {gene:get_single_search_count(gene, new_pubmed = new_pubmed) for gene in genes}## we will always put the gene in quotes
    search_term_counts = {term:get_single_search_count(term, no_search_quote = no_search_quote, new_pubmed = new_pubmed) for term in search_term}## for some special cases, we won't put the search term in quotes
    combined_counts = {}
    for gene in genes:
        for term in search_term:
            ## if one of them doesn't show up at all on it's own - they can't be found together, so we'll just short circuit it and return zero for this
            if search_term_counts[term] > 0 and gene_counts[gene] > 0:
                combined_counts[gene+term] = get_pubmed_comentions(gene, term, no_search_quote = no_search_quote, new_pubmed = new_pubmed)
            else:
                combined_counts[gene+term] = 0
    return(gene_counts,search_term_counts,combined_counts)


def get_percent_overlap(num,den):
    if min([num,den])==0:
        return(0.)
    else:
        return(num/den)


def get_comention_summary(genes, search_term, out_dir = None, no_search_quote = False, new_pubmed = False):
    gene_counts,search_term_counts,combined_counts = get_all_result_comparisons(genes, search_term, no_search_quote = no_search_quote, new_pubmed = new_pubmed)
    output = [["gene_symbol","search_term","gene_total","search_term_total","combined_total","percent_overlap_of_gene","percent_overlap_of_term","max_percent_overlap","search_url"]]
    for gene in genes:
        for term in search_term:
            gene_overlap_per = get_percent_overlap(combined_counts[gene+term],gene_counts[gene])
            term_overlap_per = get_percent_overlap(combined_counts[gene+term],search_term_counts[term])
            if new_pubmed:
                temp_url = prep_new_pubmed_url(gene, term, no_search_quote = no_search_quote)
            else:
                temp_url = prep_url(gene, term, no_search_quote = no_search_quote)
            output.append([gene,
                           term,
                           gene_counts[gene],
                           search_term_counts[term],
                           combined_counts[gene+term],
                           gene_overlap_per,
                           term_overlap_per,
                           max([gene_overlap_per,term_overlap_per]),
                           temp_url])
    if out_dir is not None:
        out_path = os.path.join(out_dir,"gene_symbol_comention_summary.tsv")
        write_table(deepcopy(output),out_path)
    return(output)


def get_comentions_from_files(gene_file, search_term_file, out_dir, no_search_quote = False):
    search_term = read_file(search_term_file, 'lines')
    genes = read_file(gene_file, 'lines')
    get_comention_summary(genes, search_term, out_dir, new_pubmed = True, no_search_quote = no_search_quote)


##############################################################

if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument("-genes", '-g',
        help="The file containing gene symbols that we'll look up",
        type = str)

    parser.add_argument("-search_terms","-search","-terms","-st", 
        help="the file with search terms that we'll look for co-mentions with the genes for",
        type = str)

    parser.add_argument("-no_search_quote",
        help="whether or not to include quotes for the search terms",
        action = 'store_true')

    parser.add_argument("-out_dir", '-out', '-o',
        help="the directory that we'll write the results to",
        type = str)

    args = parser.parse_args()

    get_comentions_from_files(args.genes, args.search_terms, args.out_dir, no_search_quote = args.no_search_quote)


#!/usr/bin/env python3
import sys, os, glob, time
import numpy as np
import random
#import scipy
from subprocess import Popen
from time import sleep
#from scipy.stats import spearmanr
#from time import time
import fileinput,pickle
from math import floor
#from scipy.stats import gaussian_kde
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib.animation import FuncAnimation
#import seaborn as sns
import h5py
from copy import deepcopy
import networkx as nx
from scipy.stats import rankdata
from scipy.interpolate import interp1d
from statsmodels.nonparametric.smoothers_lowess import lowess
import ray
import multiprocessing
try:
    from pyminer.common_functions import *
    from pyminer.pyminer_common_stats_functions import *
    from pyminer.pyminer_common_stats_functions import no_p_spear, get_Z_cutoff, fast_single_spear
    from pyminer.pyminer_get_lincs import get_cell_cycle, do_single_lincs_analysis_from_files
    from pyminer.aux.pyminer_get_lincs import get_cell_cycle, do_single_lincs_analysis_from_files
except:
    from common_functions import *
    from pyminer_common_stats_functions import *
    from pyminer_common_stats_functions import no_p_spear, get_Z_cutoff, fast_single_spear
    from pyminer_get_lincs import get_cell_cycle, do_single_lincs_analysis_from_files
    from aux.pyminer_get_lincs import get_cell_cycle, do_single_lincs_analysis_from_files
###########################################################

def get_cell_type_gene_halfway_point(mean_express_array,
                                     gene_names,
                                     gene_rho_vect,
                                     abs_rho_cutoff,
                                     cell_type_index_vect,
                                     gene_progression_dir):
    ## select the genes that are significantly above the rho cutoff
    pos_idxs = np.where(gene_rho_vect>=abs_rho_cutoff)[0]
    ## select the genes that are sinificantly below the rho cutoff
    neg_idxs = np.where(gene_rho_vect<=-1*abs_rho_cutoff)[0]
    #########
    ## get the gene expression "halfway" point. Essentially, this is 
    ## like the 'inflection point'-ish in which the expression goes up or down towards it's 'destination' expression level
    gene_mean_max = np.max(mean_express_array[:,cell_type_index_vect],axis=1)
    gene_mean_min = np.min(mean_express_array[:,cell_type_index_vect],axis=1)
    gene_halfway = (gene_mean_max+gene_mean_min)/2
    ## create a binary matrix that indicates
    greater_than_half_mean_mat = lambda i:  mean_express_array[i,:] > gene_halfway[i]
    less_than_half_mean_mat = lambda i:  mean_express_array[i,:] < gene_halfway[i]
    ## positives will increase with cell_type_index_vect
    halfway_positives = lambda i: np.argmax(greater_than_half_mean_mat(i)[cell_type_index_vect])
    ## negatives will decrease with cell_type_index_vect
    halfway_negatives = lambda i: np.argmax(less_than_half_mean_mat(i)[cell_type_index_vect])
    ## evaluate them
    pos_gene_halfway_points = [halfway_positives(idx) for idx in pos_idxs]
    neg_gene_halfway_points = [halfway_negatives(idx) for idx in neg_idxs]
    print("finished getting halfway points")
    ## make a dictionary to return the results
    pos_dict={}
    neg_dict={}
    print(gene_names[:5])
    for i in range(len(pos_idxs)):
        idx = pos_idxs[i]
        pos_dict[gene_names[idx]]=pos_gene_halfway_points[i]
    for i in range(len(neg_idxs)):
        idx = neg_idxs[i]
        neg_dict[gene_names[idx]]=neg_gene_halfway_points[i]
    up_down_dict = {"pos":pos_dict,"neg":neg_dict}
    #write_up_down_res(up_down_dict, cell_type_index_vect, gene_progression_dir)
    return(up_down_dict)


def get_all_paths_in_subgraph(sub_graph, stem_cell_id):
    ## get the leaves
    degree_dict = dict(sub_graph.degree())
    leaves = []
    for key, degree in degree_dict.items():
        if degree == 1:
            leaves.append(key)
    ## go through each leaf & get the shortest path from the stem cell 
    all_paths = []
    for leaf in leaves:
        all_paths.append(nx.shortest_path(sub_graph, source=stem_cell_id, target = leaf))
    return(all_paths)


def get_max_from_dict(in_dict,temp_set):
    temp_winner = None
    temp_win_val = -99999999
    for key in in_dict.keys():
        #print(key,":",in_dict[key])
        if key in temp_set:
            #print(temp_win_val)
            if in_dict[key]>temp_win_val:
                temp_win_val = in_dict[key]
                temp_winner = key
    return(temp_winner)


def get_all_lineage_paths(lineage_dict, stem_cell_id = None):
    ## stem cell is the one with highest page rank
    pr_dict = nx.pagerank(lineage_dict["graph"])
    commun_btwn_cent_dict = nx.communicability_betweenness_centrality(lineage_dict['graph'])
    ## maybe make a plot?
    #perhaps
    ## for each component, the the shortest path from stem cell to leaf
    comps = nx.connected_components(lineage_dict['graph'])
    all_paths = []
    for comp in comps:
        temp_nodes = deepcopy(set(comp))
        #print(temp_nodes)
        if stem_cell_id is None:
            #temp_stem = get_max_from_dict(pr_dict,temp_nodes)
            temp_stem = get_max_from_dict(commun_btwn_cent_dict,temp_nodes)
        else:
            temp_stem = stem_cell_id
        print("temp stem cell:",temp_stem)
        if temp_stem in temp_nodes:
            temp_subgraph = lineage_dict["graph"].subgraph(temp_nodes)
            all_paths += get_all_paths_in_subgraph(temp_subgraph,temp_stem)
        else:
            pass
    ## returns a list of lists with all of the paths to construct
    return(all_paths)


def get_small_distance_mat(distance_mat, exemplar_indices, scale = False):
    if type(exemplar_indices)==np.ndarray:
        exemplar_indices = exemplar_indices.tolist()
    small_dist_mat = np.zeros((np.shape(distance_mat)[0],len(exemplar_indices)))
    ## go through each of the exemplars and populate the small distance mat
    for ex in range(len(exemplar_indices)):
        ex_idx = exemplar_indices[ex]
        small_dist_mat[:,ex] = distance_mat[:,ex_idx]
    ## now, we'll change the self distance to equal the minimum non-zero distance to itself. Otherwise, exemplars will stand out like sore thumbs
    ## because realistically, if you were to actually measure the same transcriptome twise, the Euclidean distance would actually still be large 
    ## because of the random sampling. So in essense we're correcting for technical variability in measurements.
    # for ex in range(len(exemplar_indices)):
    #     ex_idx = exemplar_indices[ex]
    #     non_zero_indices = np.where(small_dist_mat[:,ex]>0)[0]
    #     small_dist_mat[ex_idx,ex] = np.min(small_dist_mat[non_zero_indices,ex])
    ## if we need to scale, then we'll scale
    if scale:
        small_dist_mat[:,ex] -= np.mean(small_dist_mat[:,ex])
        small_dist_mat[:,ex] /= np.std(small_dist_mat[:,ex])
    return(small_dist_mat)


def interpolate(x, y, n=20000, kind="cubic"):
    ## x is position on the manifold trajectory, y is the matrix of distances where each column is distance to exemplar
    new_x = (np.arange(n)/(n-1)) * np.max(x)
    interp_y = np.zeros((n,y.shape[1]))
    for i in range(y.shape[1]):
        temp_interp_func = interp1d(x, y[:,i], kind = kind)
        interp_y[:,i]=temp_interp_func(new_x)
    return(new_x, interp_y)


def lowess_smooth(full_path_x, full_path_y, frac = 0.175):
    ## first interpolate to prevent lowess from sometimes doing absolutely nothing...
    full_path_x, full_path_y = interpolate(full_path_x, full_path_y, n=full_path_x.shape[0]*2, kind="linear")
    lowess_y = np.zeros(full_path_y.shape)
    for i in range(full_path_y.shape[1]):
        temp_fit = lowess(full_path_y[:,i],full_path_x,frac=frac)
        lowess_y[:,i]=temp_fit[:,1]
    return(full_path_x,lowess_y)

#################################################

def get_local_clusters_on_path(lineage_path, current_cluster):
    if current_cluster == lineage_path[0]:
        return(lineage_path[:2])
    if current_cluster == lineage_path[-1]:
        return(lineage_path[-2:])
    if current_cluster not in lineage_path:
        print("current_cluster not in lineage_path")
    clust_loc = np.where(np.array(lineage_path)==current_cluster)[0][0]
    return(lineage_path[(clust_loc-1):(clust_loc+2)])


@ray.remote
def ray_get_closest_x_position(idx_list,
                               small_dist_mat, 
                               full_path,
                               lineage_path,
                               smooth_topology_matrix_x,
                               smooth_topology_matrix_y,
                               linear_groups):
    min_cap = np.nanmin(smooth_topology_matrix_x)
    max_cap = np.nanmax(smooth_topology_matrix_x)
    new_x_dict = {}
    for idx in idx_list:
        #### get the pertinent cell types
        ## this version of the local small_dist_mat is all cells in rows, and the ordered progression
        ## of 
        local_clust_set = set(get_local_clusters_on_path(lineage_path, linear_groups[idx]))
        group_progression = np.array(linear_groups)[full_path]
        local_clusts = []
        for i in range(len(full_path)):
            if group_progression[i] in local_clust_set:
                local_clusts.append(i)
        ## now get the Euclidean distance along this sub-space
        sum_sq_euc_dist = np.sum((small_dist_mat[idx,local_clusts] - smooth_topology_matrix_y[:,local_clusts])**2, axis = 1)
        best_x_fit = np.where(sum_sq_euc_dist == np.min(sum_sq_euc_dist))[0].tolist()[0]
        #print(best_x_fit)
        ## here is where we clip the ends of the fit, so that we only include cells that are actually along the given trajectory
        #if (best_x_fit!= 0) and (best_x_fit != np.shape(sum_sq_euc_dist)[0]):
        #    print(best_x_fit,np.shape(sum_sq_euc_dist)[0])
        if smooth_topology_matrix_x[best_x_fit] != max_cap and smooth_topology_matrix_x[best_x_fit] != min_cap:
            new_x_dict[idx] = smooth_topology_matrix_x[best_x_fit]
        elif idx in full_path:
            new_x_dict[idx] = smooth_topology_matrix_x[best_x_fit]
    return(new_x_dict)


def get_cell_embeddings_on_given_trajectory(small_dist_mat, 
                                            smooth_topology_matrix_x,
                                            smooth_topology_matrix_y,
                                            full_path,
                                            lineage_path,
                                            sample_k_lists,
                                            linear_groups,
                                            processes = None):
    if processes == None:
        processes = multiprocessing.cpu_count()
    pertinent_cell_indices = []
    for cell_type in lineage_path:
        pertinent_cell_indices += sample_k_lists[cell_type]
    index_lists = ray_get_indices_from_list(processes,pertinent_cell_indices)
    ## put the ray objects & make the calls
    ray.init()
    ray_small_dist_mat = ray.put(small_dist_mat)
    ray_smooth_topology_matrix_x = ray.put(smooth_topology_matrix_x)
    ray_smooth_topology_matrix_y = ray.put(smooth_topology_matrix_y)
    ray_calls = []
    for p in range(processes):
        ray_calls.append(ray_get_closest_x_position.remote(index_lists[p],
                                                           ray_small_dist_mat,
                                                           full_path,
                                                           lineage_path,
                                                           ray_smooth_topology_matrix_x,
                                                           ray_smooth_topology_matrix_y,
                                                           linear_groups))
    all_ray_dict_results = ray.get(ray_calls)
    ray.shutdown()
    ## start out with nans & populate for cells that are actually on the topology
    current_x_position = [np.nan]*small_dist_mat.shape[0]
    for temp_dict in all_ray_dict_results:
        for key, value in temp_dict.items():
            current_x_position[key]=value
    current_x_position = np.array(current_x_position)
    return(current_x_position)


#################################################
def are_connected_at_cutoff(temp_subgraph, pr_dict, temp_cutoff, start, end):
    temp_nodes = [key for key, value in pr_dict.items() if value>temp_cutoff]
    ## if one of them isn't in the subgraph, then they definitely can't be connected
    if start not in temp_nodes or end not in temp_nodes:
        return(False)
    ## if they are both in the graph, then we'll check that they are still connected
    temp_subgraph = temp_subgraph.subgraph(temp_nodes)
    if end in nx.node_connected_component(temp_subgraph, start):
        return(True)
    else:
        return(False)


def get_pr_cutoff(temp_subgraph, pr_dict, pr_vect, start, end):
    ## binary search for correct page rank cutoff
    # https://www.pyblog.in/programming/binary-search-in-python/
    n = len(pr_vect)
    L = 0
    R = n-1
    while L < R:
        mid = floor((L+R)/2)
        #print("L:",L)
        #print("R:",R)
        #print("mid:",mid)
        if are_connected_at_cutoff(temp_subgraph, pr_dict, pr_vect[mid], start, end):
            L = mid + 1
        elif not are_connected_at_cutoff(temp_subgraph, pr_dict, pr_vect[mid], start, end):
            R = mid - 1
        else:
            return(pr_vect[mid])
    return (pr_vect[L])


def get_minimal_graph(temp_subgraph, start, end):
    ## first filter for only points that are actually between the start and end
    print("getting minimally connected graph by page rank")
    ## find the maximum page rank cutoff at which the exemplars are still connected
    pr_dict = nx.pagerank(temp_subgraph)
    pr_vect = sorted(list(set(pr_dict.values())))
    pr_cutoff = get_pr_cutoff(temp_subgraph, pr_dict, pr_vect, start, end)
    print("current PR cutoff:",pr_cutoff)
    temp_nodes = [key for key, value in pr_dict.items() if value>=(pr_cutoff-1e-10)]
    print("start in temp_nodes:",start in temp_nodes)
    print("end in temp_nodes:",end in temp_nodes)
    return(temp_subgraph.subgraph(temp_nodes), pr_dict)


def filter_graph_for_between(temp_subgraph, start_exemplar, end_exemplar, small_dist_mat, local_clusts):
    all_nodes = sorted(list(temp_subgraph.nodes()))
    small_dist_mat = small_dist_mat[:,local_clusts]
    ## get the direction of all pertinent points to each of the two exemplars
    direction_to_start = np.array((small_dist_mat[start_exemplar,:] - small_dist_mat)>=0,dtype=int)
    direction_to_end = np.array((small_dist_mat[end_exemplar,:] - small_dist_mat)>=0,dtype=int)
    #print(direction_to_start)
    #print(direction_to_end)
    ## if they're going in the same direction, taking the difference will sum to zero.
    ## This means that the point is not between the two exemplars in Euclidean space & will therefore be removed.
    all_dims_in_same_dir_binary_mat = np.abs(direction_to_start - direction_to_end)
    #print(all_dims_in_same_dir_binary_mat)
    any_in_diff_direction = np.sum(all_dims_in_same_dir_binary_mat, axis = 1)
    #print(any_in_diff_direction)
    temp_node_idxs = set(np.where(any_in_diff_direction!=0)[0])
    original_node_set = set(all_nodes)
    temp_nodes = temp_node_idxs.intersection(original_node_set)
    #print(temp_node_idxs)
    #temp_nodes = np.array(all_nodes,dtype=int)[temp_node_idxs].tolist()
    ## add back in the exemplars. B/C of floating pt errors, sometimes they don't get included
    temp_nodes = temp_nodes.union({start_exemplar, end_exemplar})
    #print(temp_nodes)
    #print(start_exemplar in temp_nodes)
    #print(end_exemplar in temp_nodes)
    return(temp_subgraph.subgraph(temp_nodes))


def get_custom_shortest_path(temp_subgraph, start_exemplar, end_exemplar, distance_mat, small_dist_mat, start_clust, end_clust, do_minimal = True):
    if do_minimal:
        temp_subgraph = filter_graph_for_between(temp_subgraph, start_exemplar, end_exemplar, small_dist_mat, [start_clust, end_clust])
        temp_subgraph, pr_dict = get_minimal_graph(temp_subgraph, start_exemplar, end_exemplar)
    all_sp = list(nx.all_shortest_paths(temp_subgraph, source=start_exemplar, target=end_exemplar, weight = None))
    sp_distances = np.zeros(len(all_sp))
    for i in range(len(all_sp)):
        sp = all_sp[i]
        path_edges = set(zip(sp,sp[1:]))
        sp_dist = 0
        for path in path_edges:
            sp_dist+=(distance_mat[path[0],path[1]]*-1)
            #sp_dist+=(pr_dict[path[0]]*pr_dict[path[1]])
        sp_distances[i]=sp_dist
    #central_dist = np.median(sp_distances)
    #dist_to_central_dist = (sp_distances-central_dist)**2
    best_path_idx = np.where(sp_distances==np.min(sp_distances))[0]
    #best_path_idx = np.where(sp_distances==np.max(sp_distances))[0]
    #best_path_idx = np.where(dist_to_central_dist==np.min(dist_to_central_dist))[0]
    return(all_sp[best_path_idx[0]])

#################################################
#################################################
## plotting functions

def pos1_to_pos2_interpolate(pos1,pos2,steps=30, rm_nans = True):
    ## takes in two pos dicts, and returns a list of length steps with all of the updates
    all_points = []
    for key in set(pos2.keys()).intersection(set(pos1.keys())):
        num_nan1 = np.sum(np.isnan(pos1[key]))
        num_nan2 = np.sum(np.isnan(pos2[key]))
        total_nan = num_nan1+num_nan2
        #print("\t",total_nan)
        #print(pos1[key])
        #print(pos2[key])
        if total_nan > 0:
            pass
        else:
            all_points.append(key)
    dims = pos1[all_points[0]].shape
    rate_of_change = {point:np.zeros(dims) for point in all_points}
    #pos_list = [deepcopy(empty_pos) for step in range(steps)]
    for point in all_points:
        #print(point)
        #step_size = (pos2[point] - pos1[point])/steps
        rate_of_change[point] = (pos2[point] - pos1[point])/steps
        # for axis in range(pos1[point].shape[0]):
        #     #print("\t",axis)
        #     step_size = (pos2[point][axis] - pos1[point][axis])/steps
        #     #print("\t",pos1[point][axis],"to",pos2[point][axis], "in",steps,"steps")
        #     #print(step_size)
        #     if pos1[point][axis]!=pos2[point][axis]:
        #         temp_axis = crange(pos1[point][axis], pos2[point][axis]+step_size, step_size)
        #     else:
        #         temp_axis = [pos1[point][axis]]*steps
        #     #print(temp_axis)
        #     for step in range(steps):
        #         #print("\t\t",step)
        #         #print(pos_list[step][point][axis])
        #         pos_list[step][point][axis]=temp_axis[step]
    return(rate_of_change, steps)


def get_pos_from_init(init_dict,sample_k_lists, init_scale_fact = .75):
    out_pos_init = {}
    for i in range(len(sample_k_lists)):
        for cell in sample_k_lists[i]:
            out_pos_init[cell] = init_dict[i]*init_scale_fact
    return(out_pos_init)


def get_positions(subject_network, lineage_dict, sample_k_lists, transition_probability, seed = 123456):
    ## first we'll calculate the initial positions by lineage graph
    full_clust_network = nx.from_numpy_matrix(transition_probability)
    #pos_init_dict = nx.spring_layout(full_clust_network)
    pos3d_init_dict = nx.spring_layout(full_clust_network, dim=3, seed = seed)
    #full_pos_init_dict = get_pos_from_init(pos_init_dict, sample_k_lists)
    full_pos3d_init_dict = get_pos_from_init(pos3d_init_dict, sample_k_lists)
    #print("plotting 2D coords")
    #pos = linear_norm_pos_dict(nx.spring_layout(subject_network, pos = full_pos_init_dict))
    print("plotting 3D coords")
    pos3d = linear_norm_pos_dict(nx.spring_layout(subject_network, pos = full_pos3d_init_dict, dim=3, seed = seed))
    #return(pos, pos3d)
    return(None, pos3d)


# @ray.remote
# def spin_3d(fig,ax,plt,angle_set,temp_out_dir):
@ray.remote
def spin_3d(ax,angle_set,temp_out_dir):
    for angle in angle_set:
        ax.view_init(30, angle)
        out_file = os.path.join(temp_out_dir,"angle_"+str(angle)+".jpg")
        plt.savefig(out_file,
                    dpi=175,
                    bbox_inches='tight')
    return()


def plot_3d(subject_network, pos3d, expanded_color_vect, temp_out_dir, full_path=None, spin=True, processes = None):
    ## helped by: https://www.idtools.com.au/3d-network-graphs-python-mplot3d-toolkit/
    print("plotting the path in 3D")
    print(subject_network)
    angle = 0
    node_size = 2.5
    edge_width = 2.5/np.sqrt(nx.number_of_edges(subject_network))
    fig = plt.figure(figsize=(10,7))
    ax = Axes3D(fig)
    ## plot the edges
    for key, value in pos3d.items():
        if key%1000 == 0:
            print("plotting point#",key)
        xi = value[0]
        yi = value[1]
        zi = value[2]
        # Scatter plot
        dummy_point = ax.scatter(xi, yi, zi, c=[expanded_color_vect[key]], s=node_size, linewidths = 0, alpha=0.7)
    if full_path is not None:
        path_edges = set(zip(full_path,full_path[1:]))
        for i,j in enumerate(path_edges):
            #print(i,j)
            x = np.array((pos3d[j[0]][0], pos3d[j[1]][0]))
            y = np.array((pos3d[j[0]][1], pos3d[j[1]][1]))
            z = np.array((pos3d[j[0]][2], pos3d[j[1]][2]))
            # Plot the connecting lines
            #print(x,y,z)
            dummy_line = ax.plot(x, y, z, c='black', linewidth = 6)
            ## try to make the animation
            #animation = FuncAnimation(fig, func=animate_3d_spin, frames=range(360), interval=10)
            #plt.show()
            ## set the camera and save
    ax.set_axis_off()
    if processes==None:
        processes = multiprocessing.cpu_count()
    all_angles = ray_get_indices_from_list(processes,np.arange(0,360,2).tolist())
    ray_calls = []
    ray.init()
    for angle_set in all_angles:
        #ray_calls.append(spin_3d.remote(fig,ax,plt,angle_set,temp_out_dir))
        ray_calls.append(spin_3d.remote(ax,angle_set,temp_out_dir))
    ray.get(ray_calls)
    ray.shutdown()
    return(fig)

#################################################
#################################################
## fix tips
def fix_tips(subject_network, 
             start_exemplar, 
             end_exemplar, 
             small_dist_mat, 
             distance_mat,
             original_full_path, 
             lineage_path, 
             sample_k_lists,
             linear_groups):
    start_clust = lineage_path[0]
    end_clust = lineage_path[-1]
    ## first, figure out which of the points on the full path is the first point belonging to the second cluster
    linear_groups = np.array(linear_groups)
    print("\tstart_exemplar:",start_exemplar,"from clust:",linear_groups[start_exemplar])
    print("\tend_exemplar:",end_exemplar,"from clust:",linear_groups[end_exemplar])
    vector_of_cluster_membership = linear_groups[original_full_path]
    print("lineage_path",lineage_path)
    print("vector_of_cluster_membership",vector_of_cluster_membership)
    get_dist_from_pt_beginning = np.min(np.where(vector_of_cluster_membership != start_clust)[0])
    get_dist_from_pt_end = np.max(np.where(vector_of_cluster_membership != end_clust)[0])
    new_beginning_comparison_idx = original_full_path[get_dist_from_pt_beginning]
    new_end_comparison_idx = original_full_path[get_dist_from_pt_end]
    print("\tfirst pt not in first clust is pt#:",new_beginning_comparison_idx,"from clust:",linear_groups[new_beginning_comparison_idx])
    print("\tlast pt not in final clust is pt#:",new_end_comparison_idx,"from clust:",linear_groups[new_end_comparison_idx])
    ## get the new beginning and destination points
    # get the points in the first cluster, and find the one that maximizes distance from the first point in the second cluster
    points_in_current_cluster = sample_k_lists[start_clust]
    local_clusts = get_local_clusters_on_path(lineage_path, start_clust)
    sum_sq_euc_dist = np.sum((small_dist_mat[new_beginning_comparison_idx,local_clusts] - small_dist_mat[:,local_clusts])**2, axis = 1)
    #sum_sq_euc_dist += np.sum((small_dist_mat[start_exemplar,local_clusts] - small_dist_mat[:,local_clusts])**2, axis = 1)
    sum_sq_euc_dist = sum_sq_euc_dist[points_in_current_cluster]
    best_x_fit = np.where(sum_sq_euc_dist == np.max(sum_sq_euc_dist))[0].tolist()[0]
    new_beginning_pt_idx = points_in_current_cluster[best_x_fit]
    print("new_beginning_pt_idx",new_beginning_pt_idx,"from clust:",linear_groups[new_beginning_pt_idx])
    # get the points in the LAST cluster, and find the one that maximizes distance from the first point in the second cluster
    points_in_current_cluster = sample_k_lists[end_clust]
    local_clusts = get_local_clusters_on_path(lineage_path, end_clust)
    sum_sq_euc_dist = np.sum((small_dist_mat[new_end_comparison_idx,local_clusts] - small_dist_mat[:,local_clusts])**2, axis = 1)
    #sum_sq_euc_dist += np.sum((small_dist_mat[end_exemplar,local_clusts] - small_dist_mat[:,local_clusts])**2, axis = 1)
    sum_sq_euc_dist = sum_sq_euc_dist[points_in_current_cluster]
    best_x_fit = np.where(sum_sq_euc_dist == np.max(sum_sq_euc_dist))[0].tolist()[0]
    new_end_pt_idx = points_in_current_cluster[best_x_fit]
    print("new_end_pt_idx",new_end_pt_idx,"from clust:",linear_groups[new_end_pt_idx])
    ## subset the graph for the first cluster
    temp_subgraph = subject_network.subgraph(sample_k_lists[start_clust])
    new_beginning_extension = get_custom_shortest_path(temp_subgraph, start_exemplar, new_beginning_pt_idx, distance_mat, small_dist_mat, start_clust, end_clust)[1:][::-1]
    full_path = new_beginning_extension + original_full_path
    ## add the extension to the end
    temp_subgraph = subject_network.subgraph(sample_k_lists[end_clust])
    new_end_extension = get_custom_shortest_path(temp_subgraph, end_exemplar, new_end_pt_idx, distance_mat, small_dist_mat, start_clust, end_clust)[1:]
    full_path = full_path + new_end_extension
    return(full_path)


#################################################
def get_pertienent_cells_and_ordered_mat(small_dist_mat, 
                                         distance_mat, 
                                         subject_network, 
                                         lineage_path, 
                                         exemplar_indices, 
                                         sample_k_lists,
                                         processes,
                                         out_dir,
                                         linear_groups,
                                         expanded_color_vect = None,
                                         pos = None,
                                         pos3d = None):
    # temp_out_dir = os.path.join(out_dir,'_'.join(map(str,deepcopy(lineage_path))))
    # if not os.path.exists(temp_out_dir):
    #     os.makedirs(temp_out_dir)
    full_path = []
    ## get the shortest path from each exemplar
    print("getting cell embeddings along lineage")
    print("\tgetting path")
    for i in range(0,len(lineage_path)-1):
        start_clust = lineage_path[i]
        end_clust = lineage_path[i+1]
        start_exemplar = exemplar_indices[start_clust]
        end_exemplar = exemplar_indices[end_clust]
        ## first subset the embedding for only the pertinent cell types
        temp_nodes = sample_k_lists[start_clust] + sample_k_lists[end_clust]
        temp_subgraph = subject_network.subgraph(temp_nodes)
        if len(full_path)==0:
            full_path += get_custom_shortest_path(temp_subgraph, start_exemplar, end_exemplar, distance_mat, small_dist_mat, start_clust, end_clust)
            #full_path += nx.shortest_path(temp_subgraph, source=start_exemplar, target=end_exemplar, weight = None)#"weight")
            #full_path += nx.shortest_path(temp_subgraph, source=start_exemplar, target=end_exemplar, weight = "weight")
        else:
            full_path += get_custom_shortest_path(temp_subgraph, start_exemplar, end_exemplar, distance_mat, small_dist_mat, start_clust, end_clust)[1:]
                #full_path += nx.shortest_path(temp_subgraph, source=start_exemplar, target=end_exemplar, weight = None)[1:]#"weight")
                #full_path += nx.shortest_path(temp_subgraph, source=start_exemplar, target=end_exemplar, weight = "weight")[1:]
        ##################################
        ## now we fix the tips of the path
    original_full_path = deepcopy(full_path)
    full_path = fix_tips(subject_network, exemplar_indices[lineage_path[0]], exemplar_indices[lineage_path[-1]], small_dist_mat, distance_mat, full_path, lineage_path, sample_k_lists, linear_groups)
    #full_path = lineage_path
    print("\toriginal_full_path:",original_full_path)
    print("\t\tbest path was:",full_path)
    ##
    # if plot:
    #     #######################
    #     ## plot it
    #     ##
    #     node_size = 2.5
    #     edge_width = 2.5/np.sqrt(nx.number_of_edges(subject_network))
    #     path_edges = set(zip(full_path,full_path[1:]))
    #     expanded_color_vect = [cluster_dict["color_vect"][i] for i in cluster_dict["linear_groups"]]
    #     ##
    #     print("plotting 2D",subject_network,full_path)
    #     plt.clf()
    #     nx.draw(subject_network,pos,node_color=expanded_color_vect,node_size = node_size, width = edge_width)
    #     nx.draw_networkx_nodes(subject_network,pos,nodelist=full_path,node_color='r',node_size = node_size*5, width = edge_width)
    #     nx.draw_networkx_edges(subject_network,pos,edgelist=path_edges,edge_color='r',node_size = node_size, width = 10)
    #     ##
    #     plt.axis('equal')
    #     plt.show()
    # ##############################################
    ## get the pseudotime-like X position, by adding weights successively going from one node to the next
    ## the Y position that will be loess smoothed are the rows of the full small_dist_mat
    #print("check1")
    #####################################
    #####################################
    new_small_dist_mat = np.zeros((small_dist_mat.shape[0],len(full_path)))
    for i in range(len(full_path)):
        temp_full_vect = distance_mat[:,full_path[i]]
        new_small_dist_mat[:,i]=temp_full_vect
    #####################################
    #####################################
    full_path_x = np.zeros((len(full_path)))
    full_path_y = np.zeros((len(full_path),new_small_dist_mat.shape[1]))
    full_path_y[0,:]=new_small_dist_mat[full_path[0],:]
    for i in range(1,len(full_path)):
        ## first get the edge
        prev_idx = full_path[i-1]
        cur_idx = full_path[i]
        cur_x_position = distance_mat[prev_idx,cur_idx] + full_path_x[i-1]
        full_path_x[i] = cur_x_position
        full_path_y[i,:]= new_small_dist_mat[cur_idx,:]
    ## because we use negative squared euclidean distance, 
    # we'll just multiply it by -1 to bring it back into positive space, so things are more intuitive
    full_path_x *= -1
    #print("check2")
    smooth_x, smooth_y = lowess_smooth(full_path_x, full_path_y)
    #print("check3")
    new_x, new_y = interpolate(smooth_x, smooth_y)
    #new_x, new_y = interpolate(full_path_x, full_path_y)
    ## interpolate the shortest path mat to a semi-continuous topology using a loess fit in exemplar-distance space
    #new_y2 = lowess_smooth(new_x, new_y)
    # if plot:
    #     print("plotting scatters")
    #     plt.clf()
    #     plt.scatter(new_x, new_y[:,0])
    #     plt.scatter(smooth_x, smooth_y[:,0])
    #     plt.scatter(full_path_x, full_path_y[:,0])
    #     plt.show()
    print("\tgetting embeddings of all cells along lineage")
    cell_embeddings = get_cell_embeddings_on_given_trajectory(new_small_dist_mat, 
                                                              new_x,
                                                              new_y,
                                                              full_path,
                                                              lineage_path,
                                                              sample_k_lists,
                                                              linear_groups,
                                                              processes = None)
    return(cell_embeddings, full_path)


@ray.remote
def ray_cor(cell_embeddings,temp_gene, shuffle=False):
    return(fast_single_spear(cell_embeddings,temp_gene, shuffle=shuffle))

def get_embedding_correlations(cell_embeddings, full_expression, processes = None):
    if processes==None:
        processes = multiprocessing.cpu_count()
    # get the null distribution and real correlations
    null_cor_vect = np.zeros((full_expression.shape[0]))
    gene_cor_vect = np.zeros((full_expression.shape[0]))
    # i=0
    # while i < null_cor_vect.shape[0]:
    #     if i% 2000 == 0 and i!=0:
    #         print("\tworking on gene:",i)
    #     ray.init()
    #     ray_calls = []
    #     temp_idx_list = []
    #     for p in range(processes):
    #         if i < null_cor_vect.shape[0]:
    #             temp_gene = deepcopy(np.array(full_expression[i,:]))
    #             ray_calls.append(ray_cor.remote(cell_embeddings,temp_gene, shuffle=True))
    #             temp_idx_list.append(i)
    #             i+=1
    #     temp_result = ray.get(ray_calls)
    #     ray.shutdown()
    #     for i in range(len(temp_idx_list)):
    #         idx = temp_idx_list[i]
    #         gene_cor_vect[idx]=temp_result[i][0]
    #         null_cor_vect[idx]=temp_result[i][1]
    for i in range(full_expression.shape[0]):
        if i% 2000 == 0 and i!=0:
            print("\tworking on gene:",i)
            #print(null_cor_vect[i-100:i])
            #print(gene_cor_vect[i-100:i])
        #temp_gene = deepcopy(np.array(full_expression[i,:]))
        actual_res, shuffled_res = fast_single_spear(cell_embeddings,full_expression[i,:], shuffle=True)
        gene_cor_vect[i]=actual_res
        null_cor_vect[i]=shuffled_res
    cutoff = get_Z_cutoff(null_cor_vect, z = 7.5, positive = True)
    gene_cor_vect[np.isnan(gene_cor_vect)]=0
    print(np.sum(np.abs(gene_cor_vect)>cutoff), "significant correlations found")
    return(gene_cor_vect, null_cor_vect, cutoff)


###########################################################

def reverse_weights(subject_network):
    for u,v,d in subject_network.edges(data=True):
        d['weight']=(d['weight']-1)*-1
    return(subject_network)

###########################################################
def get_max_name_cor_idx(gene_names,gene_cor_vect):
    gene_idx=np.where(gene_cor_vect == np.nanmax(gene_cor_vect))[0][0]
    print("max:",gene_names[gene_idx],":",gene_cor_vect[gene_idx])
    return(gene_names[gene_idx], gene_cor_vect[gene_idx], gene_idx)


def get_min_name_cor_idx(gene_names,gene_cor_vect):
    gene_idx=np.where(gene_cor_vect == np.nanmin(gene_cor_vect))[0][0]
    print("min:",gene_names[gene_idx],":",gene_cor_vect[gene_idx])
    return(gene_names[gene_idx], gene_cor_vect[gene_idx], gene_idx)


def get_gene_cor_and_idx(gene_names, goi, gene_cor_vect):
    try:
        gene_idx = np.where(np.array(gene_names)==goi)[0][0]
    except:
        ## if we couldn't find the gene
        print("WARNING: couldn't find gene of interest:",goi)
        return(None,None)
    return(gene_cor_vect[gene_idx], gene_idx)


@ray.remote
def ray_scatter_wrapper(arg_list):
    for args in arg_list:
        ray_plot_scatter(*args)
    return()

def ray_plot_scatter(cell_embeddings,
                     gene_cor,
                     expanded_color_vect,
                     temp_gene,
                     out_dir,
                     lineage_path,
                     gene_name="",
                     do_rank = True,
                     y_jitter = True):
    fig = plt.figure()
    valid_indices = np.where(np.logical_not(np.isnan(cell_embeddings)))[0]
    ## subset the gene_subset for 
    try:
        gene_subset = temp_gene[valid_indices]
    except:
        gene_subset = temp_gene[:,valid_indices]
    #print("gene index:",gene_idx)
    if do_rank:
        x = rankdata(cell_embeddings[valid_indices], method="dense")
        y = rankdata(gene_subset,method="dense")
    else:
        x = cell_embeddings[valid_indices]
        y = gene_subset
    if y_jitter:
        y=y+np.random.random(valid_indices.shape)
    try:
        plt.scatter(x, y,
                   s=1,
                   c = expanded_color_vect[valid_indices,:])
    except:
        plt.scatter(x, y,
                   s=1,
                   c = expanded_color_vect[valid_indices])
    temp_title = str(gene_name)+'\nRho:'+str(round(gene_cor,2))
    fig.suptitle(temp_title, fontsize=15)
    path_name = "_".join(map(str,lineage_path))
    out_file = os.path.join(out_dir,path_name+"|"+gene_name+".png")
    plt.savefig(out_file,
                dpi=600,
                bbox_inches='tight')
    return


def plot_scatter(cell_embeddings,
                 gene_cor_vect,
                 cluster_dict,
                 full_expression,
                 out_dir,
                 lineage_path,
                 do_rank = True,
                 gene_names=None,
                 plot_min=False,
                 plot_max=False,
                 gene_idx=None,
                 goi = None,
                 y_jitter = True,
                 alias = None):
    fig = plt.figure()
    valid_indices = np.where(np.logical_not(np.isnan(cell_embeddings)))[0]
    expanded_color_vect = np.array([cluster_dict["color_vect"][i] for i in cluster_dict["linear_groups"]])
    #print("gene index:",gene_idx)
    if plot_min and gene_idx is None and goi is None:
        gene_idx=np.where(gene_cor_vect == np.nanmin(gene_cor_vect))[0][0]
        print("min:",gene_names[gene_idx],":",gene_cor_vect[gene_idx])
        #temp_gene = full_expression[gene_idx,:]
    #print("gene index:",gene_idx)
    if plot_max and gene_idx is None and goi is None:
        gene_idx=np.where(gene_cor_vect == np.nanmax(gene_cor_vect))[0][0]
        print("max:",gene_names[gene_idx],":",gene_cor_vect[gene_idx])
        #temp_gene = full_expression[gene_idx,:]
    #print("gene index:",gene_idx)
    if gene_idx is None and goi is not None:
        gene_idx = np.where(np.array(gene_names)==goi)[0][0]
    #print("gene index:",gene_idx)
    if gene_idx is not None:
        temp_gene = full_expression[gene_idx,:]
        gene_subset = temp_gene[valid_indices]
    else:
        gene_subset = temp_gene[:,valid_indices]
    #print("gene index:",gene_idx)
    if do_rank:
        x = rankdata(cell_embeddings[valid_indices], method="dense")
        y = rankdata(gene_subset,method="dense")
    else:
        x = cell_embeddings[valid_indices]
        y = gene_subset
    if y_jitter:
        y=y+np.random.random(valid_indices.shape)
    try:
        plt.scatter(x, y,
                   s=1,
                   c = expanded_color_vect[valid_indices,:])
    except:
        plt.scatter(x, y,
                   s=1,
                   c = expanded_color_vect[valid_indices])
    if alias is None:
        temp_title = str(gene_names[gene_idx])+'\nRho:'+str(round(gene_cor_vect[gene_idx],2))
    else:
        temp_title = str(alias)+'\nRho:'+str(round(gene_cor_vect[gene_idx],2))
    fig.suptitle(temp_title, fontsize=15)
    path_name = "_".join(map(str,lineage_path))
    out_file = os.path.join(out_dir,path_name+"|"+gene_names[gene_idx]+".png")
    plt.savefig(out_file,
                dpi=600,
                bbox_inches='tight')
    return


@ray.remote
def ray_parallel_fade(all_alpha_list, pos3d, frame_indices_list, temp_out_dir, expanded_color_vect, full_path=None):
    for current_frame in frame_indices_list:
        temp_alpha_vect = all_alpha_list[current_frame]
        print("working on frame:",current_frame)
        print(temp_alpha_vect)
        angle = 0
        node_size = 2.5
        edge_width = 2.5/np.sqrt(len(list(pos3d.keys())))
        plt.clf()
        fig = plt.figure(figsize=(10,7))
        ax = Axes3D(fig)
        ax.view_init(30, 0)
        ## plot the nodes
        for key, value in pos3d.items():
            temp_alpha = temp_alpha_vect[key]
            # if key%1000 == 0:
            #     print("plotting point#",key)
            xi = value[0]
            yi = value[1]
            zi = value[2]
            # Scatter plot
            dummy_point = ax.scatter(xi, yi, zi, c=[expanded_color_vect[key]], s=node_size, linewidths = 0, alpha=temp_alpha)
        ## plot the lines
        if full_path is not None:
            path_edges = set(zip(full_path,full_path[1:]))
            for i,j in enumerate(path_edges):
                #print(i,j)
                x = np.array((pos3d[j[0]][0], pos3d[j[1]][0]))
                y = np.array((pos3d[j[0]][1], pos3d[j[1]][1]))
                z = np.array((pos3d[j[0]][2], pos3d[j[1]][2]))
                # Plot the connecting lines
                #print(x,y,z)
                #temp_line_alpha = max([np.min(temp_alpha)/.75,.25])
                temp_line_alpha = max([np.min(temp_alpha)/.75,0])
                dummy_line = ax.plot(x, y, z, c='black', linewidth = 6, alpha = temp_line_alpha)
        ## save it
        out_file = os.path.join(temp_out_dir,"frame_"+str(current_frame)+".jpg")
        ax.set_axis_off()
        plt.savefig(out_file,
                    dpi=175,
                    bbox_inches='tight')
    return()


def one_d_vect_to_3d_pos_dict(cell_embeddings):
    pos={}
    for idx in range(cell_embeddings.shape[0]):
        pos[idx]=np.array([cell_embeddings[idx],0,0])
    return(pos)


####################
## this block from here: https://stackoverflow.com/questions/50299172/python-range-or-numpy-arange-with-end-limit-include
def cust_range(*args, rtol=1e-05, atol=1e-08, include=[True, False]):
    """
    Combines numpy.arange and numpy.isclose to mimic
    open, half-open and closed intervals.
    Avoids also floating point rounding errors as with
    >>> numpy.arange(1, 1.3, 0.1)
    array([1. , 1.1, 1.2, 1.3])

    args: [start, ]stop, [step, ]
        as in numpy.arange
    rtol, atol: floats
        floating point tolerance as in numpy.isclose
    include: boolean list-like, length 2
        if start and end point are included
    """
    # process arguments
    if len(args) == 1:
        start = 0
        stop = args[0]
        step = 1
    elif len(args) == 2:
        start, stop = args
        step = 1
    else:
        assert len(args) == 3
        start, stop, step = tuple(args)
    # determine number of segments
    n = (stop-start)/step + 1
    # do rounding for n
    if np.isclose(n, np.round(n), rtol=rtol, atol=atol):
        n = np.round(n)
    # correct for start/end is exluded
    if not include[0]:
        n -= 1
        start += step
    if not include[1]:
        n -= 1
        stop -= step
    return np.linspace(start, stop, int(n))

def crange(*args, **kwargs):
    return cust_range(*args, **kwargs, include=[True, True])

def orange(*args, **kwargs):
    return cust_range(*args, **kwargs, include=[True, False])


###################


def linear_norm_pos_dict(pos, axes = None):
    all_keys = list(pos.keys())
    dim = pos[all_keys[0]].shape
    if axes is None:
        axes = np.arange(dim[0]).tolist()
    min_vect = np.zeros(dim)
    max_vect = np.zeros(dim)
    ## get the mins and maxes
    for key, value in pos.items():
        for axis in range(dim[0]):
            min_vect[axis] = min([min_vect[axis],value[axis]])
    ## subtract the mins
    for key, value in pos.items():
        for axis in range(dim[0]):
            if axis in axes:
                temp_vect = value
                temp_vect[axis]=temp_vect[axis]-min_vect[axis]
                pos[key]=temp_vect
    ## divide by maxes
    ## get the mins and maxes
    for key, value in pos.items():
        for axis in range(dim[0]):
            max_vect[axis] = max([max_vect[axis],value[axis]])
    for key, value in pos.items():
        for axis in range(dim[0]):
            if axis in axes:
                temp_vect = value
                temp_vect[axis]=temp_vect[axis]/max_vect[axis]
                pos[key]=temp_vect
    return(pos)


def update_loc(orig_vect, rate_of_change, current_frame):
    return(orig_vect + (rate_of_change * (current_frame+1)))


@ray.remote
def unfold_subset(frame_indices_list,
                  temp_pos3d,
                  rate_of_change,
                  expanded_color_vect,
                  temp_out_dir,
                  frame_count_offset,
                  total_frames=30,
                  full_path=None,
                  alpha = 0.7):
    for current_frame in frame_indices_list:
        #temp_pos3d = pos_list[current_frame]
        print("working on frame:",current_frame)
        angle = 0
        node_size = 2.5
        edge_width = 2.5/np.sqrt(len(list(temp_pos3d.keys())))
        plt.clf()
        fig = plt.figure(figsize=(10,7))
        ax = Axes3D(fig)
        ## start out at (30,0) need to get to 270
        #angle_vect = crange(360,270,-90/len(pos_list))
        angle_vect1 = crange(30,-90,(-90-30)/(total_frames-1))
        angle_vect2 = crange(0,-90,-90/(total_frames-1))
        #ax.view_init(angle_vect[current_frame], 90)
        ax.view_init(angle_vect1[current_frame], angle_vect2[current_frame])
        #ax.view_init(-90,90)
        ## plot the nodes
        valid_points = set(rate_of_change.keys())
        for key in valid_points:
            value = temp_pos3d[key]
            #for key, value in temp_pos3d.items():
            # if key%1000 == 0:
            #     print("plotting point#",key)
            new_value = update_loc(value, rate_of_change[key], current_frame)
            xi = new_value[0]
            yi = new_value[1]
            zi = new_value[2]
            # Scatter plot
            dummy_point = ax.scatter(xi, yi, zi, c=[expanded_color_vect[key]], s=node_size, linewidths = 0, alpha=alpha)
        ## plot the lines
        if full_path is not None:
            path_edges = set(zip(full_path,full_path[1:]))
            for i,j in enumerate(path_edges):
                #print(i,j)
                start_loc = update_loc(temp_pos3d[j[0]], rate_of_change[j[0]], current_frame)
                end_loc = update_loc(temp_pos3d[j[1]], rate_of_change[j[1]], current_frame)
                x = np.array((start_loc[0], end_loc[0]))
                y = np.array((start_loc[1], end_loc[1]))
                z = np.array((start_loc[2], end_loc[2]))
                # Plot the connecting lines
                #print(x,y,z)
                dummy_line = ax.plot(x, y, z, c='black', linewidth = 6, alpha = 0.25)
        ax.set_xlim(0,1)
        ax.set_ylim(0,1)
        ax.set_zlim(0,1)
        #plt.show()
        #ax.set_axis_off()
        ## save it
        out_file = os.path.join(temp_out_dir,"frame_"+str(current_frame+frame_count_offset)+".jpg")
        plt.savefig(out_file,
                    dpi=175,
                    bbox_inches='tight')
    return()


def unfold_3d(cell_embeddings,
              pos3d,
              expanded_color_vect,
              temp_out_dir,
              full_path=None,
              fade_rate = 0.05,
              initial_alpha = 0.7,
              processes = None):
    current_frame = 0
    print("plotting the path in 3D")
    angle = 0
    node_size = 2.5
    edge_width = 2.5/np.sqrt(cell_embeddings.shape[0])
    plt.clf()
    fig = plt.figure(figsize=(10,7))
    ax = Axes3D(fig)
    ax.view_init(30, 0)
    ## plot the edges
    for key, value in pos3d.items():
        if key%1000 == 0:
            print("plotting point#",key)
        xi = value[0]
        yi = value[1]
        zi = value[2]
        # Scatter plot
        dummy_point = ax.scatter(xi, yi, zi, c=[expanded_color_vect[key]], s=node_size, linewidths = 0, alpha=0.7)
    if full_path is not None:
        path_edges = set(zip(full_path,full_path[1:]))
        for i,j in enumerate(path_edges):
            #print(i,j)
            x = np.array((pos3d[j[0]][0], pos3d[j[1]][0]))
            y = np.array((pos3d[j[0]][1], pos3d[j[1]][1]))
            z = np.array((pos3d[j[0]][2], pos3d[j[1]][2]))
            # Plot the connecting lines
            #print(x,y,z)
            dummy_line = ax.plot(x, y, z, c='black', linewidth = 6)
            ## try to make the animation
            #animation = FuncAnimation(fig, func=animate_3d_spin, frames=range(360), interval=10)
            #plt.show()
            ## set the camera and save
    ax.set_axis_off()
    ## plot the original scene
    out_file = os.path.join(temp_out_dir,"frame_"+str(current_frame)+".jpg")
    current_frame+=1
    plt.savefig(out_file,
                dpi=175,
                bbox_inches='tight')
    ## fade out the non-pertinent cells
    alpha_vect = np.ones(cell_embeddings.shape) * initial_alpha
    alpha_delta_vect = np.array(np.isnan(cell_embeddings),dtype=float)* -1 * fade_rate
    all_alpha_list = []
    while np.min(alpha_vect) > 0:
        print(np.min(alpha_vect),np.max(alpha_vect))
        alpha_vect += alpha_delta_vect
        alpha_vect[alpha_vect<0] = 0
        all_alpha_list.append(deepcopy(alpha_vect))
    ## make sure it actually went to zero - sometimes it doesn't b/c of floating point error
    alpha_vect += alpha_delta_vect
    alpha_vect[alpha_vect<0] = 0
    all_alpha_list.append(alpha_vect)
    ## now plot them
    ray.init()
    ray_calls = []
    if processes==None:
        processes = multiprocessing.cpu_count()
    ray_pos3d = ray.put(pos3d)
    frame_indices_list = get_indices(processes,len(all_alpha_list))
    for i in range(len(frame_indices_list)):
        ray_calls.append(ray_parallel_fade.remote(all_alpha_list, ray_pos3d, frame_indices_list[i], temp_out_dir, expanded_color_vect, full_path))
    ray.get(ray_calls)
    ray.shutdown()
    ## now unfold
    valid_indices = np.where(np.logical_not(np.isnan(cell_embeddings)))[0].tolist()
    rank_embeddings = deepcopy(cell_embeddings)
    rank_embeddings[valid_indices] = rankdata(cell_embeddings[valid_indices],method="dense")
    #pos_list = pos1_to_pos2_interpolate(linear_norm_pos_dict(pos3d),linear_norm_pos_dict(one_d_vect_to_3d_pos_dict(rankdata(cell_embeddings,method="dense")),[0]))
    rate_of_change, total_frames = pos1_to_pos2_interpolate(linear_norm_pos_dict(pos3d),linear_norm_pos_dict(one_d_vect_to_3d_pos_dict(rank_embeddings),[0]))
    index_lists = get_indices(processes,total_frames)
    ray_calls = []
    ray.init()
    ray_pos3d = ray.put(pos3d)
    ray_rate_of_change = ray.put(rate_of_change)
    for idx_list in index_lists:
        # unfold_subset(frame_indices_list,
        #           temp_pos3d,
        #           pos_rate_of_change,
        #           expanded_color_vect,
        #           temp_out_dir,
        #           frame_count_offset,
        #           total_frames=30,
        #           full_path=None,
        #           alpha = 0.7)
        ray_calls.append(unfold_subset.remote(idx_list,
                                              ray_pos3d,
                                              ray_rate_of_change,
                                              expanded_color_vect,
                                              temp_out_dir,
                                              len(all_alpha_list),## just for continuing the count
                                              total_frames = total_frames,
                                              full_path=full_path,
                                              alpha = 0.7))
    ray.get(ray_calls)
    ray.shutdown()
    return()


def get_continuous_lineage_pseudotimes(exemplar_indices,
                                       lineage_dict,
                                       distance_mat,
                                       subject_network,
                                       mean_express_array,
                                       sample_k_lists,
                                       full_expression,
                                       gene_names,
                                       out_dir,
                                       linear_groups,
                                       cluster_dict,
                                       stem_cell_id = None,
                                       processes = None,
                                       plot = True,
                                       goi_dict = {},
                                       seed = 123456):
    ## first get the stem cell & lineage paths for each connected component of the lineage graph
    all_lineage_paths = get_all_lineage_paths(lineage_dict, stem_cell_id=stem_cell_id)
    small_dist_mat = get_small_distance_mat(distance_mat, exemplar_indices, scale=False)
    ## invert the weights in the network so that the weighted shortest paths 
    ## correctly views it as distance rather than connection strength
    #subject_network2 = reverse_weights(subject_network)
    expanded_color_vect = [cluster_dict["color_vect"][i] for i in cluster_dict["linear_groups"]]
    lineage_path = all_lineage_paths[1]
    #lineage_path = all_lineage_paths[-1]
    all_cell_embeddings = []
    all_gene_cor = []
    all_cutoffs = []
    gene_ordering_dicts = []
    temp_out_dirs = []
    two_d_embed_scatters_dirs = []
    three_d_spin_dirs = []
    gene_progression_dirs = []
    ## now go through each lineage path
    all_results = {}
    if plot:
        pos, pos3d = get_positions(subject_network, lineage_dict, sample_k_lists, np.array(cluster_dict["transition_probability"].data), seed = seed)
        all_results["pos"]=pos
        all_results["pos3d"]=pos3d
    for lineage_path in all_lineage_paths:
        print(lineage_path)
        temp_out_dir = process_dir(os.path.join(out_dir,'_'.join(map(str,deepcopy(lineage_path)))))
        two_d_embed_scatters = process_dir(os.path.join(temp_out_dir,'pseudotime_scatters'))
        three_d_spin_dir = process_dir(os.path.join(temp_out_dir,'threeD_spin'))
        gene_progression_dir = process_dir(os.path.join(temp_out_dir,'ordered_gene_progression'))
        ## log where they are
        temp_out_dirs.append(temp_out_dir)
        two_d_embed_scatters_dirs.append(two_d_embed_scatters) 
        three_d_spin_dirs.append(three_d_spin_dir)
        gene_progression_dirs.append(gene_progression_dir)
        ## get the shortest path between all of the exemplars and other cells in the embedding matrix
        cell_embeddings, full_path = get_pertienent_cells_and_ordered_mat(small_dist_mat, 
                                                                         distance_mat, 
                                                                         subject_network, 
                                                                         lineage_path, 
                                                                         exemplar_indices, 
                                                                         sample_k_lists,
                                                                         processes,
                                                                         three_d_spin_dir,
                                                                         linear_groups)
        # len(set(cell_embeddings[np.logical_not(np.isnan(cell_embeddings))]))
        # cell_embeddings[np.logical_not(np.isnan(cell_embeddings))].shape
        # np.nanmax(cell_embeddings)
        # cell_embeddings == np.nanmax(cell_embeddings)
        # np.sum(cell_embeddings == np.nanmax(cell_embeddings))
        # np.sum(cell_embeddings == np.nanmin(cell_embeddings))
        all_cell_embeddings.append(cell_embeddings)
        ##
        ## now that we have an "X-axis" (pseudotime-like) embedding of all cells along the trajectory,
        ## subset the original expression matrix, (maybe just throw it into memmory?)
        ## 1) recalculate the null background and actual correlations of X embeddings vs Y expression
        ## 2) using that empiric FPR cutoff, and actual spearman rhos
        ## 2.1) segregate the genes into positive and negative
        ## 2.2) assign the significant genes to a cell-type level position along the X-embedding
        ##
        ## now we'll get the gene level correlations
        print("getting correlations of all genes with this trajectory")
        gene_cor_vect, null_cor_vect, cutoff = get_embedding_correlations(cell_embeddings, full_expression)
        all_gene_cor.append(gene_cor_vect)
        all_cutoffs.append(cutoff)
        gene_ordering_dicts.append(get_cell_type_gene_halfway_point(mean_express_array,
                                                                          gene_names,
                                                                          gene_cor_vect,
                                                                          cutoff,
                                                                          lineage_path,
                                                                          gene_progression_dir))
        if plot:
            if processes == 1:
                plot_scatter(cell_embeddings,gene_cor_vect,cluster_dict,full_expression,two_d_embed_scatters,lineage_path,gene_names=gene_names,plot_min=True)
                plot_scatter(cell_embeddings,gene_cor_vect,cluster_dict,full_expression,two_d_embed_scatters,lineage_path,gene_names=gene_names,plot_max=True)
                for goi, alias in goi_list.items():
                    plot_scatter(cell_embeddings,gene_cor_vect,cluster_dict,full_expression,two_d_embed_scatters,lineage_path,gene_names=gene_names,goi=goi, alias = alias)
            else:
                ray.init()
                ray_args = []
                # (cell_embeddings,
                #      gene_cor,
                #      expanded_color_vect,
                #      temp_gene,
                #      out_dir,
                #      lineage_path,
                #      gene_name="",
                #      do_rank = True,
                #      y_jitter = True)
                ## add the min and max
                # min
                expanded_color_vect = np.array([cluster_dict["color_vect"][i] for i in cluster_dict["linear_groups"]])
                name, cor, idx = get_max_name_cor_idx(gene_names,gene_cor_vect)
                ray_args.append((cell_embeddings,
                                                         cor,
                                                         expanded_color_vect,
                                                         full_expression[idx,:],
                                                         two_d_embed_scatters,
                                                         lineage_path,
                                                         name))
                # max
                name, cor, idx = get_min_name_cor_idx(gene_names,gene_cor_vect)
                ray_args.append((cell_embeddings,
                                                         cor,
                                                         expanded_color_vect,
                                                         full_expression[idx,:],
                                                         two_d_embed_scatters,
                                                         lineage_path,
                                                         name))
                ## now add the genes of interest
                for orig_id, alias in goi_dict.items():
                    cor, idx = get_gene_cor_and_idx(gene_names, orig_id, gene_cor_vect)
                    if idx is not None:
                        ray_args.append((cell_embeddings,
                                                                 cor,
                                                                 expanded_color_vect,
                                                                 full_expression[idx,:],
                                                                 two_d_embed_scatters,
                                                                 lineage_path,
                                                                 alias))
                ray_calls = ray_get_indices_from_list(processes, ray_args)
                for i in range(len(ray_calls)):
                    ray_calls[i]=ray_scatter_wrapper.remote(ray_calls[i])
                ## now 
                ray.get(ray_calls)
                ray.shutdown()
            ## now plot in 3D!
            plot_3d(subject_network, 
                    pos3d, 
                    expanded_color_vect, 
                    three_d_spin_dir, 
                    full_path=full_path,
                    processes=processes)
            unfold_3d(cell_embeddings,
                      pos3d,
                      expanded_color_vect,
                      three_d_spin_dir,
                      full_path=full_path,
                      fade_rate = 0.05,
                      initial_alpha = 0.7,
                      processes=processes)
    all_results["all_lineage_paths"]=all_lineage_paths
    all_results["all_cell_embeddings"]=all_cell_embeddings
    all_results["all_gene_cor"]=all_gene_cor
    all_results["all_cutoffs"]=all_cutoffs
    all_results["gene_ordering_dicts"]=gene_ordering_dicts
    all_results["expanded_color_vect"]=expanded_color_vect
    all_results["temp_out_dir"] = temp_out_dirs
    all_results["two_d_embed_scatters"] = two_d_embed_scatters_dirs
    all_results["three_d_spin_dir"] = three_d_spin_dirs
    all_results["gene_progression_dir"] = gene_progression_dirs
    return(all_results)


def do_continuous_lineage_analysis(infile,
                                   distance_mat_file,
                                   lineage_dict_file,
                                   cluster_dict_file,
                                   subject_network_file,
                                   mean_express_file,
                                   out_dir,
                                   hdf5=False,
                                   cols = None,
                                   gene_names = None,
                                   stem_cell_id = None,
                                   plot=True,
                                   processes = None,
                                   goi_file=None,
                                   seed=123456,
                                   human_symbol_dict_file = None):
    ###########
    lineage_dict = import_dict(lineage_dict_file)
    ###########
    if hdf5:
        row_names = read_file(gene_names,'lines')
        title = read_file(cols,'lines')
        print('reading in hdf5 file')
        h5f = h5py.File(infile, 'r')
        full_expression=h5f["infile"]
    else:
        full_expression_str = read_table(infile)
        title = full_expression_str[0]
        full_expression_np = np.array(full_expression_str)
        row_names = full_expression_np[1:,0]
        full_expression = np.array(full_expression_np[1:,1:],dtype = float)
    mean_express_array = np.array(np.array(read_table(mean_express_file))[1:,1:],dtype = float)
    #
    cell_names = title[1:]
    cluster_dict = import_dict(cluster_dict_file)
    linear_groups = cluster_dict["linear_groups"]#np.array(np.array(read_table(cluster_file))[:,1],dtype=float).tolist()
    sample_k_lists = get_sample_k_lists(linear_groups)
    ## get the exemplars
    exemplar_indices = cluster_dict["exemplar_indices"]
    ## read in the distance matrix hdf5
    dist_mat_h5 = h5py.File(distance_mat_file, 'r')
    distance_mat = dist_mat_h5["infile"]
    ## import the subject/cell network
    subject_network = import_dict(subject_network_file)
    #
    if processes == None:
        processes = multiprocessing.cpu_count()
    #
    if goi_file is not None:
        temp_goi_table = read_table(goi_file)
        goi_dict = {}
        if len(temp_goi_table[0]) == 2:
            for line in temp_goi_table:
                goi_dict[line[0]]=line[1]
        elif len(temp_goi_table[0]) == 1:
            for line in temp_goi_table:
                goi_dict[line[0]]=line[0]
        else:
            sys.exit("something wrong with GOI table.\n"+str(len(temp_goi_table[0]))+"columns")
    else:
        goi_dict = {}
    ## do the analysis after processing
    all_results = get_continuous_lineage_pseudotimes(exemplar_indices,
                                                     lineage_dict,
                                                     distance_mat,
                                                     subject_network,
                                                     mean_express_array,
                                                     sample_k_lists,
                                                     full_expression,
                                                     row_names,
                                                     out_dir,
                                                     linear_groups,
                                                     cluster_dict,
                                                     plot=plot,
                                                     stem_cell_id = stem_cell_id,
                                                     processes = processes,
                                                     goi_dict = goi_dict,
                                                     seed = 123456)
    do_lincs_lineage(all_results,human_symbol_dict_file,out_dir)
    return(all_results)


#####################################################################################

def get_susbet_from_up_down(i, cell_type_index_vect, up_dict, down_dict):
    cur_cell_type = cell_type_index_vect[i]
    up_ids = [key for key, value in up_dict.items() if value==cur_cell_type]
    down_ids = [key for key, value in down_dict.items() if value==cur_cell_type]
    #print(i,len(up_ids),"up")
    #print(i,len(down_ids),"down")
    return(up_ids, down_ids)


def write_up_down_res(up_down_dict, cell_type_index_vect, gene_progression_dir, cell_cycle_dict = None, human_symbol_dict = None, reverse = False):
    ## reorganize them to group them by the cell-type that they rise and fall on
    if reverse:
        print("REVERSING")
        cell_type_index_vect = cell_type_index_vect[::-1]
        global_down = list(up_down_dict["pos"].keys())
        global_up = list(up_down_dict["neg"].keys())
        down_dict = up_down_dict["pos"]
        up_dict = up_down_dict["neg"]
    else:
        global_up = list(up_down_dict["pos"].keys())
        global_down = list(up_down_dict["neg"].keys())
        down_dict = up_down_dict["neg"]
        up_dict = up_down_dict["pos"]
    traj_id = "start:"+"_".join(map(str,cell_type_index_vect))+":end"
    top_dir = process_dir(os.path.join(gene_progression_dir,traj_id))
    full_traj_dir = process_dir(os.path.join(top_dir, "full_trajectory"))
    #full_traj_dir = process_dir(os.path.join(gene_progression_dir, str(cell_type_index_vect[0])+"_vs_"+str(cell_type_index_vect[-1])))
    ## do the forward
    ## going from beginning to the end will be pos cor
    #pos_dir = processes_dir(os.path.join(top_dir,str(cell_type_index_vect[-1])))
    #neg_dir = processes_dir(os.path.join(top_dir,"towards_"+str(cell_type_index_vect[0])))
    all_genes_changed_on_traj = global_up + global_down
    all_up_genes = '\n'.join(list(map(str,global_up)))
    all_down_genes = '\n'.join(list(map(str,global_down)))
    ## write the global up and down along the trajectory
    fwd_file = os.path.join(full_traj_dir,str(cell_type_index_vect[-1])+'.txt')
    rev_file = os.path.join(full_traj_dir,str(cell_type_index_vect[0])+'.txt')
    make_file(all_up_genes,fwd_file)
    make_file(all_down_genes,rev_file)
    if human_symbol_dict != None:
        out_file = os.path.join(full_traj_dir,"one_shot_drugs:"+"_".join(map(str,cell_type_index_vect))+"_single_treatments.tsv")
        print(out_file)
        do_single_lincs_analysis_from_files(fwd_file,
                                            rev_file, 
                                            cell_cycle_dict, 
                                            human_symbol_dict, 
                                            out_file)
    ## now go through the step-wise results and write them
    all_stages = []
    cell_type_id = []
    all_lincs_files = [out_file]
    for i in range(len(cell_type_index_vect)):
        up_ids, down_ids = get_susbet_from_up_down(i, cell_type_index_vect, up_dict, down_dict)
        if len(up_ids)+len(down_ids) > 0:
            all_stages.append([up_ids, down_ids])
            cell_type_id.append(cell_type_index_vect[i])
    if len(all_stages)>1:
        ## if there's only one stage, then just keeping all 
        for i in range(len(all_stages)):
            stage_id = "stage_"+str(i+1)+":inflection_point_at_cluster:"+str(cell_type_id[i])
            temp_stage_dir = process_dir(os.path.join(top_dir,stage_id))
            stage_prefix = traj_id+"|"+"stage_"+str(i+1)
            temp_up_file = os.path.join(temp_stage_dir,stage_prefix+"_up.txt")
            temp_down_file = os.path.join(temp_stage_dir,stage_prefix+"_down.txt")
            up_save='\n'.join(map(str,deepcopy(all_stages[i][0])))
            down_save='\n'.join(map(str,deepcopy(all_stages[i][1])))
            make_file(up_save, temp_up_file)
            make_file(down_save, temp_down_file)
            print(stage_id)
            lincs_out_file = os.path.join(temp_stage_dir,stage_prefix+"_single_treatments.tsv")
            all_lincs_files.append(lincs_out_file)
            do_single_lincs_analysis_from_files(temp_up_file,
                                            temp_down_file, 
                                            cell_cycle_dict, 
                                            human_symbol_dict, 
                                            lincs_out_file)
    return(all_lincs_files)


def do_lincs_lineage(all_results, human_symbol_dict_file, master_out_dir):
    if type(human_symbol_dict_file)==dict:
        pass
    else:
        human_symbol_dict, human_def_dict = import_dict(human_symbol_dict_file)
    if human_symbol_dict != None:
        all_lincs_files = []
        cell_cycle_dict = get_cell_cycle(human_symbol_dict)
        for i in range(len(all_results["gene_ordering_dicts"])):
            up_down_dict = all_results["gene_ordering_dicts"][i]
            cell_type_index_vect = all_results["all_lineage_paths"][i]
            gene_progression_dir = all_results["gene_progression_dir"][i]
            all_lincs_files += write_up_down_res(up_down_dict, 
                                                      cell_type_index_vect, 
                                                      gene_progression_dir, 
                                                      cell_cycle_dict,
                                                      human_symbol_dict)
            all_lincs_files += write_up_down_res(up_down_dict, 
                                                      cell_type_index_vect, 
                                                      gene_progression_dir, 
                                                      cell_cycle_dict,
                                                      human_symbol_dict,
                                                      reverse = True)
        collate_lincs(master_out_dir, all_lincs_files)
    return()


#####################################################################################
if __name__ == "__main__":
    ###########################################################
    parser = argparse.ArgumentParser()

    parser.add_argument("-infile", '-i',
        help="the original input dataset.")

    parser.add_argument(
        '-hdf5',
        help = 'The input file is an HDF5 file',
        default = False,
        action = 'store_true')

    parser.add_argument(
        "-ID_list","-ids",
        help = 'If we are using an hdf5 file, give the row-wise IDs in this new line delimeted file',
        type = str)

    parser.add_argument(
        "-columns","-cols",
        help = 'If we are using an hdf5 file, give the column-wise IDs in this new line delimeted file',
        type = str)

    parser.add_argument("-human_symbol_dict", 
                        help="the pickled file created by pyminer_gprofiler_converter.py. This is typically 'human_orthologues.pkl'",
                        type = str)

    parser.add_argument("-distance_mat", '-dm',
                        help="the path to the full distance matrix as hdf5. Usually called 'sample_clustering_and_summary/rho_dicts/neg_euc_dist.hdf5'.")

    parser.add_argument("-lineage_dict_file", '-ld',
                        help="the lineage dictionary generated by lineage_analysis.py. This will usually be called 'lineage_dict.pkl'")

    parser.add_argument("-subject_network_file")
    
    parser.add_argument("-mean_express_file")

    parser.add_argument("-out_dir", '-o',
        help="the directory for the ouptut")

    args = parser.parse_args()
    ###########################################################
    all_results = do_continuous_lineage_analysis(args.infile,
                                             args.distance_mat_file,
                                             args.lineage_dict_file,
                                             args.cluster_dict_file,
                                             args.subject_network_file,
                                             args.mean_express_file,
                                             args.out_dir,
                                             goi_file = args.goi_file,
                                             hdf5=args.hdf5,
                                             cols = args.columns,
                                             gene_names = args.ID_list,
                                             stem_cell_id = None,
                                             plot=plot,
                                             processes = args.processes,
                                             seed = args.seed,
                                             human_symbol_dict_file = args.human_symbol_dict)
    ############################################################

initial_alpha=0.7
fade_rate=0.05
plot=False
processes = None
stem_cell_id = None
hdf5 = True
cols = "/home/scott/Documents/scRNAseq/human_breast_cancer_organoid_GSE113196/analysis/column_IDs.txt"
gene_names = "/home/scott/Documents/scRNAseq/human_breast_cancer_organoid_GSE113196/analysis/ID_list.txt"
infile = "/home/scott/Documents/scRNAseq/human_breast_cancer_organoid_GSE113196/analysis/ind4567_merged_cr1258inf_sr5011inf_log2.hdf5"
lineage_dict_file = "/home/scott/Documents/scRNAseq/human_breast_cancer_organoid_GSE113196/analysis/lineage_analysis/lineage_dict.pkl"
distance_mat_file = "/home/scott/Documents/scRNAseq/human_breast_cancer_organoid_GSE113196/analysis/sample_clustering_and_summary/rho_dicts/neg_euc_dist.hdf5"
cluster_dict_file = "/home/scott/Documents/scRNAseq/human_breast_cancer_organoid_GSE113196/analysis/sample_clustering_and_summary/clustering_plots.pkl"
subject_network_file = '/home/scott/Documents/scRNAseq/human_breast_cancer_organoid_GSE113196/analysis/sample_clustering_and_summary/cell_embedding.graphpkl'
cluster_file = "/home/scott/Documents/scRNAseq/human_breast_cancer_organoid_GSE113196/analysis/sample_clustering_and_summary/sample_k_means_groups.txt"
mean_express_file = "/home/scott/Documents/scRNAseq/human_breast_cancer_organoid_GSE113196/analysis/sample_clustering_and_summary/k_group_means.txt"
out_dir = "/home/scott/Documents/scRNAseq/human_breast_cancer_organoid_GSE113196/analysis/lineage_analysis/trajectory_plots/"
goi_file = "/home/scott/Documents/scRNAseq/human_breast_cancer_organoid_GSE113196/analysis/goi.txt"
human_symbol_dict_file = "/home/scott/Documents/scRNAseq/human_breast_cancer_organoid_GSE113196/analysis/annotations.pkl"
seed = 123456


all_results = do_continuous_lineage_analysis(infile,
                                             distance_mat_file,
                                             lineage_dict_file,
                                             cluster_dict_file,
                                             subject_network_file,
                                             mean_express_file,
                                             out_dir,
                                             goi_file = goi_file,
                                             hdf5=hdf5,
                                             cols = cols,
                                             gene_names = gene_names,
                                             stem_cell_id = None,
                                             plot=plot,
                                             processes = processes,
                                             seed = seed,
                                             human_symbol_dict_file=human_symbol_dict_file)




#!/usr/bin/env python3

##import dependency libraries
import sys,time,glob,os,pickle,fileinput,argparse
from subprocess import Popen
from operator import itemgetter
import gc, fileinput, random
import numpy as np
import argparse
from copy import deepcopy
##############################################################
##############################################################
## check that we have gprofiler installed
try:
    from gprofiler import GProfiler
except:
    sys.exit('please install gprofiler; try: pip3 install gprofiler-official')
else:
    from gprofiler import GProfiler

##############################################################
##############################################################
#import pandas as pd
##############################################################
## basic function library
def read_file(tempFile,linesOraw='lines',quiet=False):
    if not quiet:
        print('reading',tempFile)
    f=open(tempFile,'r')
    if linesOraw=='lines':
        lines=f.readlines()
        for i in range(0,len(lines)):
            lines[i]=lines[i].strip('\n')
    elif linesOraw=='raw':
        lines=f.read()
    f.close()
    return(lines)

def make_file(contents,path):
    f=open(path,'w')
    if isinstance(contents,list):
        f.writelines(contents)
    elif isinstance(contents,str):
        f.write(contents)
    f.close()

    
def flatten_2D_table(table,delim):
    #print(type(table))
    if str(type(table))=="<class 'numpy.ndarray'>":
        out=[]
        for i in range(0,len(table)):
            out.append([])
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    out[i].append(str(table[i][j]))
            out[i]=delim.join(out[i])+'\n'
        return(out)
    else:
        for i in range(0,len(table)):
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    table[i][j]=str(table[i][j])
            table[i]=delim.join(table[i])+'\n'
    #print(table[0])
        return(table)

def strip_split(line, delim = '\t'):
    return(line.strip('\n').split(delim))

def make_table(lines,delim):
    for i in range(0,len(lines)):
        lines[i]=lines[i].strip()
        lines[i]=lines[i].split(delim)
        for j in range(0,len(lines[i])):
            try:
                float(lines[i][j])
            except:
                lines[i][j]=lines[i][j].replace('"','')
            else:
                lines[i][j]=float(lines[i][j])
    return(lines)


def get_file_path(in_path):
    in_path = in_path.split('/')
    in_path = in_path[:-1]
    in_path = '/'.join(in_path)
    return(in_path+'/')


def read_table(file, sep='\t'):
    return(make_table(read_file(file,'lines'),sep))
    
def write_table(table, out_file, sep = '\t'):
    make_file(flatten_2D_table(table,sep), out_file)
    

def import_dict(f):
    f=open(f,'rb')
    d=pickle.load(f)
    f.close()
    return(d)

def save_dict(d,path):
    f=open(path,'wb')
    pickle.dump(d,f)
    f.close()

all_calls = []
def cmd(in_message, com=True):
	global all_calls
	if type(in_message)==list:
		in_message = ' '.join(in_message)
	print(in_message)
	all_calls.append(in_message)
	time.sleep(.25)
	if com:
		Popen(in_message,shell=True).communicate()
	else:
		Popen(in_message,shell=True)

def check_infile(infile):
	if os.path.isfile(infile):
		return
	else:
		sys.exit(str('could not find '+infile))

def outfile_exists(outfile):
	if os.path.isfile(outfile):
		statinfo = os.stat(outfile)
		if statinfo.st_size!=0:
			return(True)
		else:
			return(False)
	else:
		return(False)
##############################################################
##############################################################



## fix for entrez...
def process_genes(in_genes):
    out_genes = []
    for gene in in_genes:
        try:
            float(gene)
        except:
            out_genes.append(gene)
        else:
            temp_gene = str(int(float(gene)))
            out_genes.append(temp_gene)
    return(out_genes)


def do_gprofiler_analysis(infile, out, species = 'hsapiens', background = None, verbose = False, silent = True):
    ## setup the api
    gp = GProfiler('PyMINEr_'+str(random.randint(0,int(1e6))), want_header = True)

    ## import the gene names
    genes = read_file(args.infile, linesOraw = 'lines')

    if args.verbose:
        print("genes:\n",genes)

    if args.background != None:
        if os.path.isfile(args.background):
            temp_bg = read_file(args.background,linesOraw = 'lines')
        else:
            temp_bg = str(args.background)
            temp_bg = temp_bg.split(',')
        if args.verbose:
            print("Background:\n",temp_bg)
        result = gp.gprofile(process_genes(genes), custom_bg = temp_bg, organism = args.species)
    else:
        result = gp.gprofile(process_genes(genes), organism = args.species)

    if args.verbose:
        print(result)
        for line in result:
            print(line)

    if args.out == None:
        args.out = args.infile[:-4]+'_gProfiler.txt'

    write_table(deepcopy(result), args.out)
    if silent:
        return()
    else:
        return(results)


#####################################################

if __name__ == "__main__":

    parser = argparse.ArgumentParser()


    parser.add_argument("-infile",'-i', '-input', 
        help="a file containing gene ids separated on new lines",
        type = str)

    parser.add_argument("-out", '-outfile', '-o',
        help="the file that we'll write the results to",
        type = str)

    parser.add_argument("-species", '-s',
        help="a gProfiler accepted species code. Dafault = 'hsapiens'",
        type = str,
        default = 'hsapiens')

    parser.add_argument("-background", '-b',
        help="If you are going to use a custom background provide either a comma separated list (no spaces), or a file with genes on a new line each.",
        type = str,
        default = None)

    parser.add_argument(
        "-verbose","-v",
        action="store_true")

    args = parser.parse_args()

    do_gprofiler_analysis(args.infile, args.out, species = args.species, background = args.background, verbose = args.verbose, silent = True)









#!/usr/bin/env python3
##import dependency libraries
import sys,time,glob,os,pickle
from subprocess import Popen
from operator import itemgetter
import gc, fileinput
gc.enable()


## check for numpy dependency
try:
    import numpy as np
except:
    print('\n\nnumpy is a dependency for this script\nit can be installed from:\nWindows:\nhttp://www.lfd.uci.edu/~gohlke/pythonlibs/#numpy\nMac:\nhttp://sourceforge.net/projects/numpy/files/NumPy/\nLinux:\nhttps://launchpad.net/~scipy/+archive/ppa\n\n\n')
import scipy
import numpy as np
from scipy.stats import mannwhitneyu as mw
from scipy.linalg import eigh
    
from numpy import linalg
import random
from random import sample

random.seed(12345)
np.random.seed(12345)

## check for matplotlib
if '-no_graphs' in sys.argv:
    no_matplot_lib=True
    pass
else:
    print('\n\nmatplotlib is a dependency for graphs produced by this script\nit can be installed from:\nhttp://matplotlib.org/downloads.html\n\n\n')
    from matplotlib import use
    use('Agg')
    no_matplot_lib=False
    import matplotlib.pyplot as plt



######################################################################################
###############
## basic function library
def read_file(tempFile,linesOraw,quiet=False):
    if not quiet:
        print('reading',tempFile)
    f=open(tempFile,'r')
    if linesOraw=='lines':
        lines=f.readlines()
        for i in range(0,len(lines)):
            lines[i]=lines[i].strip('\n')
    elif linesOraw=='raw':
        lines=f.read()
    f.close()
    return(lines)

def make_file(contents,path):
    f=open(path,'w')
    if isinstance(contents,list):
        f.writelines(contents)
    elif isinstance(contents,str):
        f.write(contents)
    f.close()

    
def flatten_2D_table(table,delim):
    #print(type(table))
    if str(type(table))=="<class 'numpy.ndarray'>":
        out=[]
        for i in range(0,len(table)):
            out.append([])
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    out[i].append(str(table[i][j]))
            out[i]=delim.join(out[i])+'\n'
        return(out)
    else:
        for i in range(0,len(table)):
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    table[i][j]=str(table[i][j])
            table[i]=delim.join(table[i])+'\n'
    #print(table[0])
        return(table)

def strip_split(line, delim = '\t'):
    return(line.strip('\n').split(delim))

def make_table(lines,delim):
    for i in range(0,len(lines)):
        lines[i]=lines[i].strip()
        lines[i]=lines[i].split(delim)
        for j in range(0,len(lines[i])):
            try:
                float(lines[i][j])
            except:
                lines[i][j]=lines[i][j].replace('"','')
            else:
                lines[i][j]=float(lines[i][j])
    return(lines)


def read_table(file, sep='\t'):
    return(make_table(read_file(file,'lines'),sep))
    
def write_table(table, out_file, sep = '\t'):
    make_file(flatten_2D_table(table,sep), out_file)

def strip_split(line, delim = '\t'):
    return(line.strip('\n').split(delim))

def import_dict(f):
    f=open(f,'rb')
    d=pickle.load(f)
    f.close()
    return(d)

def save_dict(d,path):
    f=open(path,'wb')
    pickle.dump(d,f)
    f.close()

def cmd(in_message, com=True):
    print(in_message)
    time.sleep(.25)
    if com:
        Popen(in_message,shell=True).communicate()
    else:
        Popen(in_message,shell=True)

############################################################################################

## read in the necessary arguments
##############################################################
import argparse
parser = argparse.ArgumentParser()

parser.add_argument("-adjacency_list", "-adj","-adj_list",
    help="the adjacency list ")
parser.add_argument("-nodes_of_interest", "-noi","-nodes",
    help="the nodes that you want to subset and get their degrees of separation from")
parser.add_argument("-ID_list", "-id_list","-ids","-IDs",
    help="the list of all IDs that made the network")
parser.add_argument("-out_dir", 
    help="the directory to place the output file(s)")
parser.add_argument("-all_distances", 
    help="if you want to write a file that has all distances as a matrix. (note that the IDs will be in the same order as the input -ID_list)",
    action = 'store_true',
    default = False)
args = parser.parse_args()
##############################################################
if args.adjacency_list == None:
    sys.exit('-adjacency_list </path/to/file> is a required argument')
else:
    adj_list_file = args.adjacency_list

if args.nodes_of_interest != None:
	nodes_of_interest = args.nodes_of_interest
	nodes_of_interest = nodes_of_interest.split(',')
else:
    nodes_of_interest = None

if args.ID_list!=None:
	ID_list = read_file(args.ID_list, "lines")
	ID_hash = {}
	for i in range(0,len(ID_list)):
		ID_hash[ID_list[i]] = i
else:
	sys.exit('-ID_list </path/to/file> is a required argument')

if args.out_dir != None:
	out_dir = args.out_dir
else:
	out_dir = adj_list_file.split('/')
	out_dir = '/'.join(out_dir[:-1])+'//'


#################################################################
## first read in the adjacency list & make an adjacency matrix ##
#################################################################
print('\n\nmaking the adjacency matrix\n\n')
## make an empty adjacency matrix
G_dense = np.zeros((len(ID_list),len(ID_list)),dtype = bool)

## go through the adjacency list and populate the adjacnecy matrix
first_line = True
for line in fileinput.input(adj_list_file):
    #print(first_line)
    if not first_line:
        temp_line = strip_split(line)
        idx1 = ID_hash[temp_line[0]]
        idx2 = ID_hash[temp_line[1]]
        G_dense[idx1,idx2] = 1
        G_dense[idx2,idx1] = 1
    else:
        first_line = False

################################################################
########## get all shortest paths from & to all nodes ##########
################################################################
print('\n\nfinding all of the shortest paths\n\n')
G_sparse = scipy.sparse.csr_matrix(G_dense)
short_path_mat = scipy.sparse.csgraph.shortest_path(G_sparse)

if args.all_distances:
    temp_outfile = out_dir+'/all_shortest_paths_matrix.txt'
    write_table(short_path_mat,temp_outfile)
else:
    pass



## subset the distance matrix to get all distances from the nodes of interest
header = ['genes']
out_array = np.transpose(np.array([ID_list]))

if nodes_of_interest != None:
    for n in nodes_of_interest:
        if n not in ID_hash:
            sys.exit(str('could not find ' + n + ' in the adjacency matrix'))
        else:
            header.append('dist_to_'+str(n))
            ## get the distances
            temp_dist_vect = short_path_mat[:,ID_hash[n]]

            ## make the shortest list for output
            ## [all_nodes, current_node_of_interest, shortest_path]
            #out_list = np.hstack(out_array, np.transpose(np.array([[n]*len(ID_list)])) ) )

            #print(np.shape(out_list))
            #print(np.shape(temp_dist_vect))
            out_array = np.hstack((out_array, np.transpose(np.array([temp_dist_vect]))) )

            #temp_outfile = out_dir+'/'+n+'_shortest_path_list.txt'
            #print('writing output shortest paths to:\n',temp_outfile)

out_array = np.vstack((np.array([header]), out_array))

temp_outfile = out_dir+'/'+'genes_of_interest_shortest_path_list.txt'
write_table(out_array,temp_outfile)
    		
    		
		








#################################################################




















